"""Script that tests how much time it takes to send an empty (null) CRTP
packet on the radio to a given address.
"""

import sys

from aiocflib.crtp.crtpstack import CRTPPacket
from aiocflib.drivers.crazyradio import Crazyradio, RadioConfiguration
from aiocflib.utils import timing


async def run_test():
    # General results:
    #
    # Sending a CRTP packet to an online target takes 1.06 + 0.02*k msec, where
    # k is the number of bytes in the packet.
    #
    # Sending a CRTP packet to an offline target that does not respond takes
    # 3.433 + 0.033*k msec, where k is the number of bytes in the packet.

    URI = "radio://0/80/2M/E7E7E7E701"
    URI_MISSING = "radio://0/80/2M/E7E7E7E7FA"

    device = await Crazyradio.from_uri(URI)
    config_existing = RadioConfiguration.from_uri(URI)
    config_missing = RadioConfiguration.from_uri(URI_MISSING)

    async with device as radio:
        for size in (1, 5, 10, 15, 20, 25, 30):
            print(
                f"Packet size: {size} byte(s)...", end="", file=sys.stderr, flush=True
            )
            row = []
            for config in (config_existing, config_missing):
                async with radio.configure(config):
                    packet = CRTPPacket.null().to_bytes()
                    if len(packet) < size:
                        packet += b"\x00" * (size - len(packet))
                    with timing() as t:
                        for _i in range(5000):
                            await radio.send_and_receive_bytes(packet)
                    row.append(t.elapsed / 5)
            row = "\t".join(map(str, row))
            print(" done.", file=sys.stderr, flush=True)
            print(f"{size}\t{row}")


if __name__ == "__main__":
    import anyio

    anyio.run(run_test)



"""Classes related to accessing the logging subsystem of a Crazyflie."""

from anyio import Lock
from contextlib import asynccontextmanager, AsyncExitStack
from dataclasses import dataclass
from enum import IntEnum
from functools import partial
from itertools import count, zip_longest
from struct import Struct, error as StructError
from typing import (
    Any,
    AsyncIterable,
    Awaitable,
    Callable,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Tuple,
    Union,
)

from aiocflib.crtp import CRTPPacket, CRTPPort
from aiocflib.errors import error_to_string
from aiocflib.utils import anop
from aiocflib.utils.concurrency import aclosing
from aiocflib.utils.toc import TOCCache, fetch_table_of_contents_gracefully
from aiocflib.utils.typing import Disposer

from .crazyflie import Crazyflie

__all__ = ("Log", "LogBlock")

#: The maximum size of a CRTP packet payload
MAX_LOG_DATA_PACKET_SIZE = 28


class LoggingChannel(IntEnum):
    """Enum representing the names of the channels of the logging service in
    the CRTP protocol.
    """

    TABLE_OF_CONTENTS = 0
    CONTROL = 1
    DATA = 2


class LoggingTOCCommand(IntEnum):
    """Enum representing the names of the table-of-contents commands in the
    logging service of the CRTP protocol.

    These commands are valid for LoggingChannel.TABLE_OF_CONTENTS (i.e.
    channel 0).
    """

    GET_ITEM = 0
    GET_INFO = 1
    GET_ITEM_V2 = 2
    GET_INFO_V2 = 3


class LoggingControlCommand(IntEnum):
    """Enum representing the names of the control commands in the logging
    service of the CRTP protocol.

    These commands are valid for LoggingChannel.CONTROL (i.e. channel 1).
    """

    CREATE_BLOCK = 0
    APPEND_BLOCK = 1
    DELETE_BLOCK = 2
    START_LOGGING = 3
    STOP_LOGGING = 4
    RESET = 5
    CREATE_BLOCK_V2 = 6
    APPEND_BLOCK_V2 = 7


#: Dictionary mapping integer type codes to their C types, Python structs and
#: aliases
_type_properties = {
    # C type, Python struct, aliases
    0x01: ("uint8_t", Struct("<B"), ("uint8", "u8")),
    0x02: ("uint16_t", Struct("<H"), ("uint16", "u16")),
    0x03: ("uint32_t", Struct("<L"), ("uint32", "u32")),
    0x04: ("int8_t", Struct("<b"), ("int8", "i8")),
    0x05: ("int16_t", Struct("<h"), ("int16", "i16")),
    0x06: ("int32_t", Struct("<i"), ("int32", "i32")),
    0x07: ("float", Struct("<f"), ()),
    0x08: ("fp16", Struct("<h"), ()),
}


class VariableType(IntEnum):
    """Enum containing the possible types of a log variable and the corresponding
    numeric identifiers.
    """

    UINT8 = 1
    UINT16 = 2
    UINT32 = 3
    INT8 = 4
    INT16 = 5
    INT32 = 6
    FLOAT = 7
    FP16 = 8

    @classmethod
    def to_type(cls, value):
        """Converts an integer, string or VariableType_ instance to a
        VariableType.
        """
        if isinstance(value, cls):
            return value
        elif isinstance(value, str):
            return _type_names.get(value)
        else:
            return VariableType(value)

    @property
    def aliases(self) -> Tuple[str]:
        """Returns the registered type aliases of this type."""
        return (_type_properties[self][0],) + _type_properties[self][2]

    @property
    def length(self) -> int:
        """Returns the number of bytes that a single value of this log
        variable would occupy.
        """
        return self.struct.size

    @property
    def struct(self) -> Struct:
        """Returns a Python struct that can be used to encode or decode
        log variables of this type.
        """
        return _type_properties[self][1]

    def encode_value(self, value) -> bytes:
        """Encodes a single value of this log variable type into its raw
        byte-level representation.
        """
        return self.struct.pack(value)


#: Type specification for objects that can be converted into a log variable type
VariableTypeLike = Union[str, int, VariableType]

#: Dictionary mapping string type aliases to types
_type_names = {alias: type for type in VariableType for alias in type.aliases}


@dataclass(frozen=True)
class VariableSpecification:
    """Class representing the specification of a single log variable of the
    Crazyflie."""

    id: int
    type: int
    group: str
    name: str

    @classmethod
    def from_bytes(cls, data: bytes, id: int):
        try:
            type = data[0] & 0x0F
            group, name, *rest = data[1:].split(b"\x00")
            return cls(
                id=id,
                type=type,
                group=group.decode("ascii"),
                name=name.decode("ascii"),
            )
        except Exception:
            raise ValueError("invalid log variable description") from None

    def encode_value(self, value) -> bytes:
        """Encodes a single value of this log variable into its raw byte-level
        representation.
        """
        return _type_properties[self.type][1].pack(value)

    @property
    def full_name(self) -> str:
        """Returns the fully-qualified name of the log variable, which is
        the concatenation of the group and the name of the variable, separated
        by a dot.
        """
        return "{0.group}.{0.name}".format(self)

    def parse_value(self, data: bytes):
        """Parses the raw byte-level representation of a single value of this
        log variable, as received from the Crazyflie, and returns the
        corresponding Python value.
        """
        return _type_properties[self.type][1].unpack(data)[0]

    def to_bytes(self) -> bytes:
        header = int(self.type) & 0x0F

        parts = []
        parts.append(bytes((header,)))
        parts.append(self.group.encode("ascii"))
        parts.append(b"\x00")
        parts.append(self.name.encode("ascii"))
        parts.append(b"\x00")

        return b"".join(parts)


@dataclass(frozen=True)
class LogBlockItem:
    """A single item in a log block specification that holds the name of a
    log variable and the type it should be fetched as over the wire from the
    Crazyflie. The type does not need to match the type that is used to _store_
    the same variable in the Crazyflie firmware - conversion will occur
    on-the-fly.
    """

    name: str
    id: int
    fetch_as: VariableType
    stored_as: VariableType

    def to_bytes(self) -> bytes:
        """Returns a byte-level representation of this item that can be used in
        a `CREATE_BLOCK` request.
        """
        return bytes(
            (
                ((self.fetch_as << 4) & 0xF0) | (self.stored_as & 0x0F),
                self.id & 0xFF,
                self.id >> 8,
            )
        )


@dataclass(frozen=True)
class LogMessage:
    """Value object representing a single log message from the Crazyflie."""

    timestamp: int
    block: "LogBlock"
    items: Tuple
    handler: Optional["LogMessageHandler"]

    @classmethod
    def from_bytes(
        cls,
        data: bytes,
        block: "LogBlock",
        handler: Optional["LogMessageHandler"] = None,
    ):
        return cls(
            block=block,
            timestamp=int.from_bytes(data[1:4], byteorder="little"),
            items=block._decode_values(data),
            handler=handler,
        )

    def process(self, *args, **kwds):
        """Syntactic sugar for calling the message handler associated with the
        message, witn the message as its first argument.

        Any additional arguments are forwarded intact to the handler.
        """
        if self.handler:
            return self.handler(self, *args, **kwds)

    def to_dict(self) -> Dict[str, Any]:
        """Converts the items in the log message to a dictionary based on the
        log items in the associated block.
        """
        return self.block._to_dict(self.items)


#: Type specification for log message handlers in a log session
LogMessageHandler = Callable[[LogMessage], Union[None, Awaitable[None]]]


def _process_period_and_frequency(
    *,
    period: Optional[float] = None,
    frequency: Optional[float] = None,
    period_msec: Optional[int] = None,
    default: int = 100,
) -> int:
    """Processes `period` and `frequency` keyword arguments from a
    function call, returning an appropriate period in milliseconds.

    Parameters:
        period: the logging period, in seconds
        period_msec: the logging period, in milliseconds. Takes precedence
            over `period` or `frequency`.
        frequency: the logging frequency, in Hertz. Takes precedence over
            `period` if both are given
        default: the default logging period to return, in milliseconds,
            if neither the period nor the frequency are given

    Returns:
        the logging period, in milliseconds
    """
    if period_msec is not None:
        return int(period_msec)
    if frequency is None:
        frequency = 1.0 / period if period is not None else 10.0
    return int(1000.0 // frequency)


class LogBlock:
    """Specification of a single log block that bundles together a desired
    logging period and a list of variables to log.
    """

    _owner: "Log"

    _id: Optional[int]
    _items: List[LogBlockItem]
    _struct: Optional[Struct]

    def __init__(self, owner: "Log"):
        """Constructor.

        Parameters:
            owner: the log object that owns this specification
        """
        self._owner = owner
        self._disposer = None

        self._id = None
        self._items = []
        self._struct = None

    @property
    def id(self) -> Optional[int]:
        """The numeric ID of the block when it is already submitted to the
        Crazyflie, or `None` if it is not submitted yet.
        """
        return self._id

    @id.setter
    def id(self, value: Optional[int]) -> None:
        if self._id == value:
            return

        if self._id is not None and value is not None:
            raise ValueError("block already has a different ID")

        self._id = value

    @property
    def is_submitted(self) -> bool:
        """Returns whether this log specification has already been submitted
        to the Crazyflie.
        """
        return self._disposer is not None

    @property
    def items(self) -> Iterable[LogBlockItem]:
        """Iterates over the items in this log block."""
        return iter(self._items)

    @property
    def packet_size(self) -> int:
        """Returns the total number of bytes that the data in this log
        configuration would occupy.
        """
        return sum(item.fetch_as.length for item in self._items)

    def add_variable(
        self, name: str, type: Optional[Union[int, VariableType]] = None
    ) -> None:
        """Adds a new variable to this logging block."""
        toc = self._owner._variables_by_name
        try:
            spec = toc[name]
        except KeyError:
            raise KeyError("no such variable in log TOC: {0!r}".format(name)) from None

        if type is None:
            type = spec.type

        item = LogBlockItem(
            name=name,
            id=spec.id,
            stored_as=VariableType(spec.type),
            fetch_as=VariableType(type),
        )
        self._items.append(item)

    async def receive(
        self,
        *,
        period: Optional[float] = None,
        period_msec: Optional[int] = None,
        frequency: Optional[float] = None,
    ) -> AsyncIterable[LogMessage]:
        """Async generator that yields log messages according to the
        specification from the Crazyflie.

        This function will automatically submit the log specification to the
        Crazyflie if it hasn't been submitted yet. The generator will remove
        the log specification from the Crazyflie when the generator is closed
        if and only if the log specification was submitted by the generator
        itself.

        Parameters:
            period: the logging period, in seconds
            period_msec: the logging period, in milliseconds. Takes precedence
                over `period` or `frequency`.
            frequency: the logging frequency, in Hertz. Takes precedence over
                `period` if both are given
        """
        start_args = {
            "period": period,
            "period_msec": period_msec,
            "frequency": frequency,
        }
        async with self.submitted(allow_multi=True):
            async with self.started(**start_args):
                async for packet in self._owner.data_packets():
                    yield LogMessage.from_bytes(packet.data, block=self)

    async def start(
        self,
        *,
        period: Optional[float] = None,
        period_msec: Optional[int] = None,
        frequency: Optional[float] = None,
    ) -> Disposer:
        """Starts this log specification on the Crazyflie.

        Also submits the log specification to the Crazyflie if it has not been
        submitted yet.

        This is a low-level function; typically you should use the `started()`
        context manager to ensure that everything is cleaned up properly.

        Parameters:
            period: the logging period, in seconds
            period_msec: the logging period, in milliseconds. Takes precedence
                over `period` or `frequency`.
            frequency: the logging frequency, in Hertz. Takes precedence over
                `period` if both are given

        Returns:
            a function that can be used to stop the log block on the Crazyflie
        """
        if self.id is None:
            raise RuntimeError("log block was not submitted yet")

        period_msec = _process_period_and_frequency(
            period=period, period_msec=period_msec, frequency=frequency
        )
        await self._owner._start_log_block_by_id(self.id, period_msec)
        return partial(self._owner._stop_log_block_by_id, self.id)

    @asynccontextmanager
    async def started(
        self,
        *,
        period: Optional[float] = None,
        period_msec: Optional[int] = None,
        frequency: Optional[float] = None,
    ):
        """Asynchronous context manager that starts the log block with the given
        logging period upon entering the context and stops it upon exiting.

        Also submits the log specification to the Crazyflie if it has not been
        submitted yet.

        Parameters:
            period: the logging period, in seconds
            period_msec: the logging period, in milliseconds. Takes precedence
                over `period` or `frequency`.
            frequency: the logging frequency, in Hertz. Takes precedence over
                `period` if both are given
        """
        disposer = await self.start(
            period=period, period_msec=period_msec, frequency=frequency
        )
        try:
            yield
        finally:
            await disposer()

    async def submit(self, allow_multi: bool = False) -> Disposer:
        """Submits this log specification to the Crazyflie and creates a new
        log block.

        This is a low-level function; typically you should use the `submitted()`
        context manager to ensure that everything is cleaned up properly.

        Parameters:
            allow_multi: whether to allow multiple submissions. When the block
                is already submitted and this argument is `True`, the function
                will do nothing and return an empty disposer function. Otherwise
                the function will raise a RuntimeError when trying to submit
                the same block multiple times

        Returns:
            a function that can be used to remove the log block from the
            Crazyflie
        """
        if self.is_submitted:
            if allow_multi:
                return anop
            else:
                raise RuntimeError("log block is already submitted to Crazyflie")

        id, self._disposer = await self._owner._submit_block(self)
        self.id = id

        format_strings = [item.fetch_as.struct.format[1:] for item in self._items]
        if format_strings and isinstance(format_strings[0], bytes):
            # Python <3.7
            format_strings = [fmt.decode("ascii") for fmt in format_strings]  # type: ignore

        self._struct = Struct("<" + "".join(format_strings))

        return self._dispose

    @asynccontextmanager
    async def submitted(self, allow_multi: bool = False):
        """Async context manager that submits the log specification to the
        Crazyflie when entering the context and removes it when exiting the
        context.
        """
        disposer = await self.submit()
        try:
            yield
        finally:
            await disposer()

    def to_bytes(self) -> bytes:
        """Returns a byte-level representation of this logging block that can
        be used in a `CREATE_BLOCK` request.
        """
        self._validate_packet_size()
        return b"".join(item.to_bytes() for item in self._items)

    def _decode_values(self, data: bytes) -> Tuple:
        """Decodes the values received in a log message from the Crazyflie.

        Parameters:
            data: the data part of the Crazyflie log message, starting from the
                log block ID.

        Returns:
            the decoded values
        """
        return self._struct.unpack_from(data, offset=4)  # type: ignore

    async def _dispose(self):
        """Removes this log block from the Crazyflie."""
        if self.id is not None:
            self.id = None

        if self._disposer is not None:
            try:
                await self._disposer()
            except RuntimeError:
                # Well, maybe the Crazyflie was disconnected?
                pass
            self._disposer = None

    def _to_dict(self, values: Iterable[Any]) -> Dict[str, Any]:
        return {
            item.name: value
            for item, value in zip_longest(self._items, values, fillvalue=None)
            if item is not None
        }

    def _validate_packet_size(self):
        """Checks whether the contents of this log specification would fit into
        a single CRTP packet.

        Raises:
            ValueError: if there are too many entries in the specification
        """
        size = self.packet_size
        if size > MAX_LOG_DATA_PACKET_SIZE:
            raise ValueError(
                "log packet too large ({0} bytes, max is {1})".format(
                    size, MAX_LOG_DATA_PACKET_SIZE
                )
            )


class LogSession:
    """Class representing a single logging session that consists of multiple
    log blocks, log block frequency settings and handler functions.

    The session can be used either as an asynchronous generator (via its
    `messages()` method), which will yield the decoded log messages
    corresponding to the log blocks registered in the session, or directly as
    an asynchronous context manager. In both cases, the log blocks are submitted
    to the Crazyflie and then started when the iteration starts or when the
    execution enters the context, and everything is cleaned up when the
    iteration stops or when the execution leaves the context.
    """

    def __init__(self, owner: "Log"):
        """Constructor.

        Parameters:
            owner: the log object that owns this specification
        """
        self._owner = owner
        self._blocks = []

        self._exit_stack = None
        self._id_mapping = None

        self._remove_existing_log_blocks_when_starting = False
        self._cleanup_gracefully = False

    def add_block(
        self,
        block: LogBlock,
        *,
        period: Optional[float] = None,
        period_msec: Optional[int] = None,
        frequency: Optional[float] = None,
        handler: Optional[LogMessageHandler] = None,
    ) -> None:
        """Adds a new block to the session.

        The block must not be registerd in any other session and it must not
        be submitted to a Crazyflie yet.

        Removing blocks is not supported yet; if you need to remove a block,
        close the session and start a new one.

        Parameters:
            block: the log block to add

        Keyword arguments:
            period: the logging period, in seconds
            period_msec: the logging period, in milliseconds. Takes precedence
                over `period` or `frequency`.
            frequency: the logging frequency, in Hertz. Takes precedence over
                `period` if both are given
            handler: the handler function to call when log data related to this
                block arrives in the session. The handler is called only when
                the log session is used as a context manager. When the log
                session is used as an async generator, the handler is attached
                to the generated log message in the `handler` property instead,
                and the caller can decide whether the handler should be called
                or not.
        """
        if block.is_submitted:
            raise RuntimeError("block is already submitted to a Crazyflie")
        if block in self._blocks:
            raise RuntimeError("block is already added to this session")

        period_msec = _process_period_and_frequency(
            period=period, period_msec=period_msec, frequency=frequency
        )
        self._blocks.append((block, period_msec, handler))

    def configure(
        self,
        *,
        graceful_cleanup: Optional[bool] = None,
        remove_existing_log_blocks: Optional[bool] = None,
    ) -> None:
        """Conifigures the behaviour of the log session during startup and
        cleanup.

        Parameters:
            graceful_cleanup: whether to attempt to clean up gracefully when the
                session is stopped. Exceptions will be handled and silenced
                during cleanup when possible.
            remove_existing_log_blocks: whether to remove all existing log
                blocks from the Crazyflie when this session starts.
        """
        if graceful_cleanup is not None:
            self._cleanup_gracefully = bool(graceful_cleanup)
        if remove_existing_log_blocks is not None:
            self._remove_existing_log_blocks_when_starting = bool(
                remove_existing_log_blocks
            )

    def create_block(
        self,
        *args,
        period: Optional[float] = None,
        period_msec: Optional[int] = None,
        frequency: Optional[float] = None,
        handler: Optional[LogMessageHandler] = None,
    ) -> LogBlock:
        """Shorthand function for creating a log block and adding it immediately
        to the session.

        Each positional argument must be the name of a variable to add to the
        log block.

        Keyword arguments:
            period: the logging period, in seconds
            period_msec: the logging period, in milliseconds. Takes precedence
                over `period` or `frequency`.
            frequency: the logging frequency, in Hertz. Takes precedence over
                `period` if both are given
            handler: the handler function to call when log data related to this
                block arrives in the session. The handler is called only when
                the log session is used as a context manager. When the log
                session is used as an async generator, the handler is attached
                to the generated log message in the `handler` property instead,
                and the caller can decide whether the handler should be called
                or not.
        """
        block = self._owner.create_block()
        for arg in args:
            block.add_variable(arg)
        self.add_block(
            block,
            period=period,
            period_msec=period_msec,
            frequency=frequency,
            handler=handler,
        )
        return block

    async def __aenter__(self):
        if self._exit_stack is not None:
            raise RuntimeError("cannot use the same session twice")

        self._exit_stack = AsyncExitStack()
        await self._exit_stack.__aenter__()

        if self._remove_existing_log_blocks_when_starting:
            await self._owner.reset()

        id_mapping = {}
        for entry in self._blocks:
            block, _, _ = entry
            await self._exit_stack.enter_async_context(block.submitted())
            id_mapping[block.id] = entry

        for block, period_msec, _ in self._blocks:
            await self._exit_stack.enter_async_context(
                block.started(period_msec=period_msec)
            )

        self._id_mapping = id_mapping

    async def __aexit__(self, exc_type, exc, tb) -> bool:
        try:
            assert self._exit_stack is not None
            return await self._exit_stack.__aexit__(exc_type, exc, tb)
        except Exception as ex:
            if not self._cleanup_gracefully:
                raise ex
            return True
        finally:
            self._exit_stack = None
            self._id_mapping = None

    async def messages(self) -> AsyncIterable[LogMessage]:
        """Yields log messages matching the log blocks specified in the session,
        _without_ calling the handler functions registered on them. You can
        call the handler functions by invoking the `process()` method on the
        yielded log messages.
        """
        if self._id_mapping is None:
            raise RuntimeError(
                "You must enter the session context first; use 'async with'"
            )

        id_mapping = self._id_mapping
        async for packet in self._owner.data_packets():
            log_block_id = packet.data[0]
            entry = id_mapping.get(log_block_id)
            if entry is not None:
                block, _, handler = entry
                yield LogMessage.from_bytes(packet.data, block=block, handler=handler)

    async def process_messages(self) -> None:
        """Async task that processes log messages matching the log blocks
        specified in the session, calling the appropriate log handler function
        for each of them. The handler functions must be synchronous; if you
        want to spawn a long-running task in response to a message, spawn it
        inside the handler function on your own or send the message to a queue,
        which can then be processed from another task.
        """
        async with aclosing(self.messages()) as gen:
            async for message in gen:
                message.process()


class Log:
    """Class representing the handler of messages related to the logging
    subsystem of a Crazyflie instance.
    """

    _block_id_generator: Iterator[int]
    _cache: Optional[TOCCache]
    _crazyflie: Crazyflie
    _operation_lock: Lock

    _variables: List[VariableSpecification]
    _variables_by_name: Dict[str, VariableSpecification]

    def __init__(self, crazyflie: Crazyflie):
        """Constructor.

        Parameters:
            crazyflie: the Crazyflie for which we need to handle the parameter
                subsystem related messages
        """
        self._cache = crazyflie._get_cache_for("log_toc")
        self._crazyflie = crazyflie

        self._block_id_generator = count()

        self._variables = None  # type: ignore
        self._variables_by_name = None  # type: ignore

        self._operation_lock = Lock()

    def create_block(self) -> LogBlock:
        """Creates a new, empty log block specification object that can be
        used to start logging variables from the Crazyflie.

        If you plan to use multiple log blocks at the same time, consider
        using `create_session()` instead.
        """
        return LogBlock(self)

    def create_session(self) -> LogSession:
        """Creates a new logging session object that can be used to start
        multiple log blocks at the same time and automatically call handler
        functions associated to them.
        """
        return LogSession(self)

    async def data_packets(self) -> AsyncIterable[CRTPPacket]:
        """Async generator that yields logging-related messages from a
        Crazyflie.

        This includes messages related to TOC handling and log control packets,
        not only log data. If you are interested in log data packets only,
        use ``data_packets()``.
        """
        async for packet in self._crazyflie.packets(port=CRTPPort.LOGGING):
            if packet.channel == LoggingChannel.DATA and len(packet.data) >= 4:
                yield packet

    async def packets(self) -> AsyncIterable[CRTPPacket]:
        """Async generator that yields logging-related messages from a
        Crazyflie.

        This includes messages related to TOC handling and log control packets,
        not only log data. If you are interested in log data packets only,
        use ``data_packets()``.
        """
        async for packet in self._crazyflie.packets(port=CRTPPort.LOGGING):
            yield packet

    async def reset(self):
        """Resets the logging framework and clears all logging blocks."""
        async with self._operation_lock:
            await self._crazyflie.run_command(
                port=CRTPPort.LOGGING,
                channel=LoggingChannel.CONTROL,
                command=LoggingControlCommand.RESET,
            )
        self._block_id_generator = count()

    async def validate(self):
        """Ensures that the basic information about the parameters of the
        Crazyflie are downloaded, and that the log subsystem is in a known
        state with no log blocks.
        """
        if self._variables is not None:
            return

        self._variables, self._variables_by_name = await self._validate()

    async def _create_log_block(self, id: int, block: LogBlock) -> None:
        """Creates a new log block with the given ID on the Crazyflie."""
        await self.validate()
        async with self._operation_lock:
            response = await self._crazyflie.run_command(
                port=CRTPPort.LOGGING,
                channel=LoggingChannel.CONTROL,
                command=(LoggingControlCommand.CREATE_BLOCK_V2, id),
                data=block.to_bytes(),
            )

        status = int(response[0])
        if status:
            raise RuntimeError(
                "Log block creation request returned error code {0} ({1})".format(
                    status, error_to_string(status)
                )
            )

    async def _delete_log_block_by_id(self, id: int) -> None:
        """Deletes the log block with the given ID from the Crazyflie."""
        async with self._operation_lock:
            response = await self._crazyflie.run_command(
                port=CRTPPort.LOGGING,
                channel=LoggingChannel.CONTROL,
                command=(LoggingControlCommand.DELETE_BLOCK, id),
            )

        status = int(response[0])
        if status:
            raise RuntimeError(
                "Log block deletion request returned error code {0} ({1})".format(
                    status, error_to_string(status)
                )
            )

    async def _get_log_variable_spec_by_index(
        self, index: int
    ) -> VariableSpecification:
        """Returns the specification of the logging variable with the given
        index.

        Parameters:
            index: the index of the logging variable to retrieve

        Returns:
            the specification of the logging variable with the given index
        """
        response = await self._crazyflie.run_command(
            port=CRTPPort.LOGGING,
            channel=LoggingChannel.TABLE_OF_CONTENTS,
            command=(LoggingTOCCommand.GET_ITEM_V2, index & 0xFF, index >> 8),
        )
        if not response:
            raise IndexError("parameter index out of range")
        return VariableSpecification.from_bytes(response, id=index)

    async def _get_table_of_contents_info(self) -> Tuple[int, int]:
        """Returns basic information about the table of contents of the
        log variable list, including the number of log variables and the CRC32
        hash of the parameter table.

        Returns:
            the number of parameters and the CRC32 hash of the parameter table
        """
        response = await self._crazyflie.run_command(
            port=CRTPPort.LOGGING,
            channel=LoggingChannel.TABLE_OF_CONTENTS,
            command=LoggingTOCCommand.GET_INFO_V2,
        )
        try:
            length, hash, _, _ = Struct("<HIBB").unpack(response)
        except StructError:
            raise ValueError("invalid logging TOC info response") from None

        return length, hash

    async def _start_log_block_by_id(self, id: int, period_msec: int) -> None:
        """Starts streaming messages from the log block with the given ID."""
        period_byte = int(period_msec / 10)
        if period_byte < 0 or period_byte > 255:
            raise ValueError("logging period must be between 0 and 2.55 seconds")

        async with self._operation_lock:
            response = await self._crazyflie.run_command(
                port=CRTPPort.LOGGING,
                channel=LoggingChannel.CONTROL,
                command=(LoggingControlCommand.START_LOGGING, id),
                data=(period_byte,),
            )

        status = int(response[0])
        if status:
            raise RuntimeError(
                "Log block start request returned error code {0} ({1})".format(
                    status, error_to_string(status)
                )
            )

    async def _stop_log_block_by_id(self, id: int) -> None:
        """Stops streaming messages from the log block with the given ID."""
        async with self._operation_lock:
            response = await self._crazyflie.run_command(
                port=CRTPPort.LOGGING,
                channel=LoggingChannel.CONTROL,
                command=(LoggingControlCommand.STOP_LOGGING, id),
            )

        status = int(response[0])
        if status:
            raise RuntimeError(
                "Log block stop request returned error code {0} ({1})".format(
                    status, error_to_string(status)
                )
            )

    async def _submit_block(self, block: LogBlock) -> Tuple[int, Disposer]:
        """Submits a log block to the Crazyflie for registration.

        Returns:
            the ID of the registered log block and an async function that can be
            called to remove the log block from the Crazyflie
        """
        id = next(self._block_id_generator)
        await self._create_log_block(id, block)
        return id, partial(self._delete_log_block_by_id, id)

    async def _validate(self):
        """Downloads the basic information about the logging subsystem of the
        Crazyflie, and that the log subsystem is in a known
        state with no log blocks.
        """
        # TODO(ntamas): when connecting to multiple drones with the same TOC,
        # fetch the log items only from one of them
        parameters = await fetch_table_of_contents_gracefully(
            self._cache,
            self._get_table_of_contents_info,
            self._get_log_variable_spec_by_index,
            VariableSpecification.from_bytes,
            VariableSpecification.to_bytes,
        )
        by_name = {parameter.full_name: parameter for parameter in parameters}

        await self.reset()

        return parameters, by_name



from __future__ import annotations

from anyio import to_thread

from contextlib import asynccontextmanager
from typing import Callable, List, Optional, TYPE_CHECKING

from aiocflib.crtp.crtpstack import CRTPPacket
from aiocflib.utils.addressing import CrazyradioAddress, parse_radio_uri
from aiocflib.utils.concurrency import ObservableValue

from .base import CRTPDriver
from .registry import register

if TYPE_CHECKING:
    import cflinkcpp


__all__ = ("CppRadioDriver",)


def import_cflinkcpp():
    """Lazily imports the `cflinkcpp` module that provides the native C++
    extension to communicate with a Crazyflie over the radio or USB.
    """
    try:
        import cflinkcpp

        return cflinkcpp
    except ImportError:
        raise RuntimeError(  # noqa: B904
            "You need to install the 'cflinkcpp' module to use the native C++-based radio driver"
        )


@register("cppradio")
class CppRadioDriver(CRTPDriver):
    """CRTP driver that allows us to communicate with a Crazyflie via a
    Crazyradio dongle using a native C++-based driver.

    Attributes:
        polling_strategy: a callable that decides how often we should poll the
            downlink if there are no packets that we want to send to the
            Crazyflie
        resending_strategy: a callable that decides whether we should resend the
            last packet if it failed or whether we should drop the connection
    """

    _connection: "cflinkcpp.Connection"
    _packet_factory: Callable[[], "cflinkcpp.Packet"]

    @asynccontextmanager
    async def _connected_to(self, uri: str):
        # strip middleware and stuff from the URI
        scheme, sep, rest = uri.partition("://")
        if scheme.startswith("usb"):
            uri = "usb://" + rest
        else:
            uri = "radio://" + rest

        cflinkcpp = import_cflinkcpp()
        connection = await to_thread.run_sync(cflinkcpp.Connection, uri)

        try:
            self._connection = connection
            self._packet_factory = cflinkcpp.Packet
            yield self
        finally:
            self._packet_factory = None  # type: ignore
            self._connection = None  # type: ignore
            # TODO(ntamas): why does connection.close() never return when
            # called on a worker thread???
            # await to_thread.run_sync(connection.close)
            connection.close()

    def __init__(self):
        """Constructor."""
        self._connection = None  # type: ignore
        self._link_quality = ObservableValue(0.0)
        self._packet_factory = None  # type: ignore

    @property
    def address(self) -> Optional[CrazyradioAddress]:
        """The address that the driver will be configured for, or ``None`` if
        the driver has no URI.
        """
        if not self._uri:
            return None

        try:
            config = parse_radio_uri(self._uri)
        except Exception:
            return None

        return config["address"] if config else None

    @property
    def is_safe(self) -> bool:
        # TODO(ntamas): make this configurable
        return True

    @property
    def link_quality(self) -> ObservableValue[float]:
        return self._link_quality

    @property
    def name(self) -> str:
        return "cppradio"

    async def notify_rebooted(self) -> None:
        # TODO(ntamas): does the C++ driver restore safe link mode after a
        # reboot?
        pass

    async def receive_packet(self) -> CRTPPacket:
        """Receives a single CRTP packet.

        Returns:
            the next CRTP packet that was received
        """
        while True:
            native_packet = await to_thread.run_sync(self._connection.recv, 100)
            if native_packet.valid:
                packet = CRTPPacket()
                packet.port = native_packet.port
                packet.channel = native_packet.channel
                packet.data = native_packet.payload
                return packet

    async def send_packet(self, packet: CRTPPacket) -> None:
        """Sends a CRTP packet.

        Parameters:
            packet: the packet to send
        """
        native_packet = self._packet_factory()
        native_packet.port = packet.port
        native_packet.channel = packet.channel
        native_packet.payload = bytes(packet.data)
        await to_thread.run_sync(self._connection.send, native_packet)

    @classmethod
    async def scan_interfaces(
        cls, address: Optional[CrazyradioAddress] = None
    ) -> List[str]:
        """Scans all interfaces of this type for available Crazyflie quadcopters
        and returns a list with appropriate connection URIs that could be used
        to connect to them.

        Parameters:
            address: the address of the Crazyflie to look for; `None` means to
                use the default address

        Returns:
            the list of connection URIs where a Crazyflie was detected; an empty
            list is returned for interfaces that do not support scanning
        """
        if isinstance(address, bytes):
            address_as_int = int.from_bytes(address, "big", signed=False)
        else:
            address_as_int = None

        scan = import_cflinkcpp().Connection.scan
        uris = scan(address_as_int) if address_as_int is not None else scan()
        result = []
        for uri in uris:
            if uri.startswith("radio://"):
                uri = "cppradio://" + uri[len("radio://") :]
            result.append(uri)

        return result

    async def use_safe_link(self) -> None:
        """Instructs the driver to start using safe-link mode to ensure
        guaranteed packet delivery to the remote peer.
        """
        # TODO(ntamas): implement this!
        pass



from __future__ import annotations

from anyio import sleep
from binascii import hexlify
from typing import Optional, TYPE_CHECKING

from aiocflib.bootloader.types import BootloaderCommand
from aiocflib.bootloader.target import BootloaderTargetType
from aiocflib.crtp import (
    CRTPDispatcher,
    CRTPDevice,
    CRTPDriver,
    CRTPPort,
    LinkControlChannel,
)
from aiocflib.errors import TimeoutError
from aiocflib.utils.concurrency import ObservableValue
from aiocflib.utils.toc import TOCCache, TOCCacheLike

__all__ = ("Crazyflie",)

if TYPE_CHECKING:
    from .commander import Commander
    from .console import Console
    from .high_level_commander import HighLevelCommander
    from .led_ring import LEDRing
    from .lighthouse import Lighthouse
    from .localization import Localization
    from .log import Log
    from .mem import Memory
    from .motors import Motors
    from .param import Parameters
    from .platform import Platform


class Crazyflie(CRTPDevice):
    """Objects representing a single Crazyflie device.

    This object should be used as a context manager; the methods of this object
    that communicate with the Crazyflie must only be called within the context
    established by the instance, e.g.::

        async with Crazyflie(uri) as cf:
            # ...do anything with the Crazyflie here...
            pass
        # Connection to the Crazyflie closes when the context is exited
    """

    @classmethod
    async def is_responding_at(cls, uri: str, *, tries: int = 1) -> bool:
        """Returns whether a Crazyflie drone is responding at the given URI.

        Parameters:
            uri: the URI to try to connect to
            tries: specifies how many times we should try to connect to the
                radio URI
        """
        instance = cls(uri)

        while tries > 0:
            try:
                async with instance:
                    while tries > 0:
                        try:
                            name = await instance.platform.get_device_type_name()
                            if name is None or not name:
                                return False
                            else:
                                return name.startswith("Crazyflie")
                        except TimeoutError:
                            tries -= 1
            except Exception:
                tries -= 1

        return False

    def __init__(self, uri: str, cache: Optional[TOCCacheLike] = None):
        """Constructor.

        Creates a Crazyflie_ instance from a URI specification.

        Parameters:
            uri: the URI where the Crazyflie can be reached
        """
        super().__init__(uri)

        self._cache = TOCCache.create(cache) if cache else None

        # Initialize sub-modules; avoid circular import
        from .app_channel import AppChannel
        from .commander import Commander
        from .console import Console
        from .high_level_commander import HighLevelCommander
        from .led_ring import LEDRing
        from .lighthouse import Lighthouse
        from .localization import Localization
        from .log import Log
        from .mem import Memory
        from .motors import Motors
        from .param import Parameters
        from .platform import Platform

        self._app_channel = AppChannel(self)
        self._commander = Commander(self)
        self._console = Console(self)
        self._high_level_commander = HighLevelCommander(self)
        self._led_ring = LEDRing(self)
        self._lighthouse = Lighthouse(self)
        self._localization = Localization(self)
        self._log = Log(self)
        self._memory = Memory(self)
        self._motors = Motors(self)
        self._parameters = Parameters(self)
        self._platform = Platform(self)

    def _get_cache_for(self, namespace: str) -> Optional[TOCCache]:
        """Returns a namespaced TOC cache instance to be used by submodules
        for caching data, or `None` if the Crazyflie instance was constructed
        without a cache.
        """
        return self._cache.namespace(namespace) if self._cache else None

    async def _prepare_link(self, driver: CRTPDriver) -> None:
        await driver.use_safe_link()

    @property
    def commander(self) -> Commander:
        """The low-level (roll-pitch-yaw-thrust) commander module of the
        Crazyflie.
        """
        return self._commander

    @property
    def console(self) -> Console:
        """The console message handler module of the Crazyflie."""
        return self._console

    @property
    def dispatcher(self) -> CRTPDispatcher:
        """Returns the packet dispatcher that dispatches incoming messages to
        the appropriate handler functions.

        You may then use the dispatcher to register handler functions to the
        messages you are interested in.
        """
        return self._dispatcher

    @property
    def high_level_commander(self) -> HighLevelCommander:
        """The high-level commander module of the Crazyflie."""
        return self._high_level_commander

    @property
    def led_ring(self) -> LEDRing:
        """The LED ring of the Crazyflie."""
        return self._led_ring

    @property
    def link_quality(self) -> ObservableValue[float]:
        """Returns an observable link quality measure from the underlying
        link.
        """
        return (
            self._driver.link_quality if self._driver else ObservableValue.constant(0.0)
        )

    @property
    def lighthouse(self) -> Lighthouse:
        """The Lighthouse subsystem of the Crazyflie."""
        return self._lighthouse

    @property
    def localization(self) -> Localization:
        """The localization subsystem of the Crazyflie."""
        return self._localization

    @property
    def log(self) -> Log:
        """The logging subsystem of the Crazyflie."""
        return self._log

    @property
    def mem(self) -> Memory:
        """The memory subsystem of the Crazyflie. This is a compatibility alias
        of ``self.memory`` for sake of compatibility with the official
        Crazyflie library.
        """
        return self._memory

    @property
    def memory(self) -> Memory:
        """The memory subsystem of the Crazyflie."""
        return self._memory

    @property
    def motors(self) -> Motors:
        """The motors subsystem of the Crazyflie."""
        return self._motors

    @property
    def param(self) -> Parameters:
        """The parameters subsystem of the Crazyflie. This is a compatibility
        alias of ``self.parameters`` for sake of compatibility with the official
        Crazyflie library.
        """
        return self._parameters

    @property
    def parameters(self) -> Parameters:
        """The parameters subsystem of the Crazyflie."""
        return self._parameters

    @property
    def platform(self) -> Platform:
        """The platform-related message handler module of the Crazyflie."""
        return self._platform

    @property
    def uri(self):
        """The URI where the Crazyflie resides."""
        return self._uri

    async def _reboot(self, to_bootloader: bool = False) -> bytes:
        """Implementation of the common parts of ``reboot()`` and
        ``reboot_to_bootloader()``.
        """
        # Initiate the reset and wait for the acknowledgment
        response = await self.run_command(
            port=CRTPPort.LINK_CONTROL,
            channel=LinkControlChannel.BOOTLOADER,
            command=(BootloaderTargetType.NRF51, BootloaderCommand.RESET_INIT),
            timeout=1,
            attempts=5,
        )

        # Store the new address from the response -- this will be used if we
        # are rebooting to bootloader mode
        address = b"\xb1" + response[3::-1]

        # Acknowledgment received, now we can send the reset command.
        await self.send_packet(
            port=CRTPPort.LINK_CONTROL,
            channel=LinkControlChannel.BOOTLOADER,
            data=(
                BootloaderTargetType.NRF51,
                BootloaderCommand.RESET,
                0 if to_bootloader else 1,
            ),
        )

        return address

    async def reboot(self, to_bootloader: bool = False) -> None:
        """Sends a packet to the Crazyflie that reboots its main processor."""
        await self._reboot()
        if self._driver:
            await self._driver.notify_rebooted()

        # Give some time for the outbound thread to send the packet
        await sleep(0.1)

    async def reboot_to_bootloader(self) -> str:
        """Sends a packet to the Crazyflie that reboots its main processor
        into its bootloader.

        It is advised that you close the connection context to the Crazyflie
        after this operation.

        Note that when warm-booting to the bootloader, the Crazyflie will use
        a different address (constructed from the first four bytes of its CPU
        ID and a fixed prefix) than the one that was used while in firmware
        mode. This function will return a new radio-based connection URI that
        can be used to re-connect to the bootloader.

        Returns:
            a new connection URI that can be used to re-connect to the bootloader
        """
        new_address_bytes = await self._reboot(to_bootloader=True)

        # Give some time for the outbound thread to send the packet
        await sleep(0.1)

        # Construct the new URI
        new_address = hexlify(new_address_bytes).decode("ascii").upper()
        scheme, _, _ = self.uri.partition("://")
        if not scheme.startswith("radio"):
            scheme = "radio"
        return "{0}://0/0/2M/{1}".format(scheme, new_address)

    async def resume(self) -> None:
        """Sends a packet to the Crazyflie that wakes up its main processor
        from a suspended state.
        """
        await self.send_packet(
            port=CRTPPort.LINK_CONTROL,
            channel=LinkControlChannel.BOOTLOADER,
            data=(BootloaderTargetType.NRF51, BootloaderCommand.RESUME),
        )

        # Give some time for the outbound thread to send the packet
        await sleep(0.1)

        # Notify the driver that the Crazyflie was rebooted
        if self._driver:
            await self._driver.notify_rebooted()

    async def suspend(self) -> None:
        """Sends a packet to the Crazyflie that suspends its main processor.
        You can wake up the Crazyflie later with ``self.resume()`` or with the
        power button.

        You are advised to close the Crazyflie context after executing this
        method and re-establish it later. This is to ensure that the safe-link
        mode is restored properly after the Crazyflie wakes up.
        """
        await self.send_packet(
            port=CRTPPort.LINK_CONTROL,
            channel=LinkControlChannel.BOOTLOADER,
            data=(BootloaderTargetType.NRF51, BootloaderCommand.SUSPEND),
        )

        # Give some time for the outbound thread to send the packet
        await sleep(0.1)

    async def shutdown(self) -> None:
        """Sends a packet to the Crazyflie that turns it off completely."""
        await self.send_packet(
            port=CRTPPort.LINK_CONTROL,
            channel=LinkControlChannel.BOOTLOADER,
            data=(BootloaderTargetType.NRF51, BootloaderCommand.SHUTDOWN),
        )

        # Give some time for the outbound thread to send the packet
        await sleep(0.1)


async def test():
    from aiocflib.crtp import MemoryType
    from aiocflib.utils import timing

    logging = False

    # uri = "cppradio://0/80/2M/E7E7E7E701"
    uri = "radio://0/80/2M/E7E7E7E701"
    # uri = "sitl://"
    # uri = "usb://0"

    if logging:
        uri = uri.replace("://", "+log://")

    async with Crazyflie(uri, cache="/tmp/cfcache") as cf:
        print("Firmware version:", await cf.platform.get_firmware_version())
        print("Protocol version:", await cf.platform.get_protocol_version())
        print("Device type:", await cf.platform.get_device_type_name())

        with timing("Fetching log TOC"):
            await cf.log.validate()
        with timing("Fetching parameters TOC"):
            await cf.parameters.validate()

        async with cf.parameters.set_and_restore("ring.effect", 6, 0):
            await sleep(1)

        with timing("Reading from memory"):
            memory = await cf.memory.find(MemoryType.LED)
            data = b"\xfc\x00" * 8
            await memory.write(0, data)
            await memory.read(0, len(data))

        with timing("Writing to memory with checksum"):
            await cf.memory.write_with_checksum(
                MemoryType.LED, 0, b"\xde\xad\xbe\xef", only_if_changed=True
            )
            await cf.memory.write_with_checksum(
                MemoryType.LED, 0, b"\xde\xad\xbe\xef", only_if_changed=True
            )

        await cf.led_ring.flash()

        async with cf.parameters.set_and_restore("ring.effect", 6, 0):
            persistence_state = await cf.parameters.get_persistence_state("ring.effect")
            if persistence_state.is_stored:
                await cf.parameters.clear_persisted_value("ring.effect")

            persistence_state = await cf.parameters.get_persistence_state("ring.effect")
            if persistence_state.is_stored:
                raise RuntimeError("ring.effect should not be persisted")

            await cf.parameters.persist("ring.effect")

            persistence_state = await cf.parameters.get_persistence_state("ring.effect")
            if not persistence_state.is_stored:
                raise RuntimeError("ring.effect should be persisted now")
            if persistence_state.stored_value != 6:
                raise RuntimeError("ring.effect persisted value should be 6")

            await cf.parameters.clear_persisted_value("ring.effect")
            persistence_state = await cf.parameters.get_persistence_state("ring.effect")
            if persistence_state.is_stored:
                raise RuntimeError("ring.effect should not be persisted")

        """
        session = cf.log.create_session()
        session.create_block(
            "pm.vbat",
            "stateEstimate.x",
            "stateEstimate.y",
            "stateEstimate.z",
            frequency=5,
        )

        async with session:
            with move_on_after(3):
                await session.process_messages()
        """


if __name__ == "__main__":
    from aiocflib.crtp import init_drivers
    import trio

    init_drivers()
    try:
        trio.run(test)
    except KeyboardInterrupt:
        pass



from os import strerror
from typing import Optional, Union

__all__ = ("error_to_string",)


class NotFoundError(RuntimeError):
    """Error thrown when a detection / scanning routine failed."""

    pass


class TimeoutError(RuntimeError):
    """Error thrown when a command sent to a CRTP-based device (for instance,
    a Crazyflie) has timed out.
    """

    pass


class CRTPCommandError(RuntimeError):
    """Error thrown when a command sent to a CRTP-based device (for instance,
    a Crazyflie) returned an error code.
    """

    def __init__(self, message: Optional[str] = None, code: int = 0):
        message = message or f"CRTP command returned error {code}"
        super().__init__(message)


def error_to_string(value: Union[int, bytes]) -> str:
    """Converts a Crazyflie error code returned in some CRTP packets to its
    human-readable description.

    Parameters:
        value: the error code, either as an integer or as a bytes object of
            length 1
    """
    if isinstance(value, bytes):
        if len(value) == 1:
            value = value[0]
        else:
            return "not an error code"
    try:
        # Crazyflie uses standard POSIX error codes
        return strerror(value)
    except ValueError:
        return "unknown error"



"""Functions and classes related to color handling."""

from colour import Color as ColorBase
from typing import Tuple, Union

__all__ = ("Color", "ColorLike", "to_color")


#: Typing for objects that can be converted into an RGB color
ColorLike = Union[str, Tuple[float, float, float]]


class Color(ColorBase):
    def get_intrgb(self):
        return tuple(max(0, min(255, round(x * 255))) for x in self.rgb)

    def set_intrgb(self, value):
        value = tuple(x / 255.0 for x in value)
        self.set_rgb(value)

    def get_rgb888(self):
        r, g, b = self.get_intrgb()
        return (r << 16) | (g << 8) | b

    def set_rgb888(self, value):
        r, g, b = ((value >> 16) & 0xFF), ((value >> 8) & 0xFF), (value & 0xFF)
        self.set_intrgb(r, g, b)


def to_color(value: ColorLike) -> Color:
    """Converts a string or an RGB tuple into a color object.

    RGB tuples must specify the components in tn [0; 255] range.
    """
    if not isinstance(value, str):
        r, g, b = value
        return Color(rgb=(r / 255.0, g / 255.0, b / 255.0))
    else:
        return Color(value)



"""Script that tests how much time it takes to send an empty (null) CRTP
packet on the radio to a given address.
"""

from aiocflib.crazyflie import Crazyflie
from aiocflib.crtp import CRTPPort, LinkControlChannel
from aiocflib.crtp.drivers import init_drivers
from aiocflib.utils import timing


async def run_test():
    URI = "cppradio://0/80/2M/E7E7E7E701"

    PACKET_SIZE = 24
    NUM_PACKETS = 1024

    init_drivers()

    async with Crazyflie(URI, cache="/tmp/cfcache") as cf:
        total_bytes = 0
        with timing() as t:
            for i in range(NUM_PACKETS):
                data = bytes([i & 0xFF] * PACKET_SIZE)
                response = await cf.run_command(
                    port=CRTPPort.LINK_CONTROL,
                    channel=LinkControlChannel.ECHO,
                    data=data,
                )
                assert data == response
                total_bytes += len(response)

        print(
            f"{total_bytes} bytes, {t.elapsed:.2f} seconds, {total_bytes / t.elapsed:.3f} bytes/sec"
        )


if __name__ == "__main__":
    import anyio

    anyio.run(run_test)



from aiocflib.crtp import CRTPDevice, CRTPPort, LinkControlChannel
from aiocflib.errors import NotFoundError
from aiocflib.drivers.crazyradio import Crazyradio
from aiocflib.utils.addressing import BootloaderAddressSpace
from anyio import sleep
from typing import List, Optional, Union

from .target import BootloaderTarget, BootloaderTargetType
from .types import BootloaderCommand, BootloaderProtocolVersion

__all__ = ("Bootloader",)


class Bootloader(CRTPDevice):
    """Objects representing a single Crazyflie device when it is in bootloader
    mode.

    This object should be used as a context manager; the methods of this object
    that communicate with the bootloader must only be called within the context
    established by the instance, e.g.::

        async with Bootloader(uri) as cf:
            # ...do anything with the bootloader here...
            pass
        # Connection to the bootloader closes when the context is exited
    """

    _targets: Optional[List[BootloaderTarget]]

    @classmethod
    async def detect_all(cls) -> List[str]:
        """Uses all connected Crazyradio dongles to scan for available Crazyflie
        quadcopters that are in bootloader mode, and returns a list with
        appropriate connection URIs that could be used to connect to them.

        TODO(ntamas): can this be used to find Crazyflies that are in bootloader
        mode after a warm (not cold) reboot?

        Returns:
            the list of connection URIs where a bootloader was detected
        """
        devices = await Crazyradio.detect_all()

        results: List[str] = []
        for index, device in enumerate(devices):
            address_space = BootloaderAddressSpace(index=index)
            async with device as radio:
                items = await radio.scan(address_space)
                results.extend(item.to_uri(index) for item in items)

        return results

    @classmethod
    async def detect_one(cls, *, tries: int = 1) -> str:
        """Uses all connected Crazyradio dongles to scan for available Crazyflie
        quadcopters that are in bootloader mode, finds the first one and returns
        the URI that can be used to connect to it.

        Parameters:
            tries: specifies how many times we should try to connect to a single
                radio URI

        Returns:
            the URI of a single Crazyflie in bootloader mode
        """
        devices = await Crazyradio.detect_all()

        for index, _device in enumerate(devices):
            address_space = BootloaderAddressSpace(index=index)
            for uri in address_space:
                found = await cls.is_responding_at(uri, tries=tries)
                if found:
                    return uri

        raise NotFoundError()

    @classmethod
    async def is_responding_at(cls, uri: str, *, tries: int = 1) -> bool:
        """Returns whether the bootloader is responding at the given radio URI.

        Parameters:
            uri: the radio URI, typically ``radio://X/0`` or ``radio://X/110``,
                where X is the index of the Crazyradio device
            tries: specifies how many times we should try to connect to the
                radio URI
        """
        device = await Crazyradio.from_uri(uri)

        async with device as radio:
            while tries > 0:
                result = await radio.scan([uri])
                if result:
                    return True
                tries -= 1

        return False

    def __init__(self, uri: str):
        """Constructor.

        Creates a Bootloader_ instance from a URI specification.

        Parameters:
            uri: the URI where the bootloader can be reached
        """
        super().__init__(uri)

        self._targets = None

    async def find_target(
        self, type: Union[BootloaderTargetType, str]
    ) -> BootloaderTarget:
        """Finds the first bootloader target of the given type.

        Raises:
            NotFoundError: if there is no such bootloader target
        """
        type = BootloaderTargetType.from_string(type)
        targets = await self.get_targets()
        for target in targets:
            if target.id == type:
                return target
        raise NotFoundError("no such bootloader target")

    async def get_targets(self) -> List[BootloaderTarget]:
        """Returns information about the possible bootloader targets. Loads it
        from the bootloader if necessary.
        """
        await self.validate()
        assert self._targets is not None
        return self._targets

    async def _reboot(self, to_firmware: bool = False) -> None:
        """Implementation of the common parts of ``reboot()`` and
        ``reboot_to_firmware()``.
        """
        # Initiate the reset and wait for the acknowledgment
        await self.run_bootloader_command(
            command=(BootloaderTargetType.NRF51, BootloaderCommand.RESET_INIT)
        )

        # Acknowledgment received, now we can send the reset command.
        await self.send_bootloader_packet(
            data=(
                BootloaderTargetType.NRF51,
                BootloaderCommand.RESET,
                1 if to_firmware else 0,
            )
        )

    async def reboot(self, to_firmware: bool = False) -> None:
        """Sends a packet to the bootloader that reboots it."""
        await self._reboot()

        # Give some time for the outbound thread to send the packet
        await sleep(0.1)

    async def reboot_to_firmware(self) -> None:
        """Sends a packet to the bootloader that reboots the main processor
        of the device into the uploaded firmware.

        It is advised that you close the connection context to the bootloader
        after this operation.
        """
        await self._reboot(to_firmware=True)

        # Give some time for the outbound thread to send the packet
        await sleep(0.1)

    async def run_bootloader_command(self, **kwds) -> bytes:
        """Shortcut to ``self.run_command()`` with the CRTP port and channel
        set to the values that make sure that the bootloader will process the
        command.

        The bootloader responds only to packets sent on the CRTP link control
        port, channel 3.
        """
        new_kwds = {"timeout": 1, "attempts": 5}
        new_kwds.update(kwds)
        return await self.run_command(
            port=CRTPPort.LINK_CONTROL,
            channel=LinkControlChannel.BOOTLOADER,
            **new_kwds
        )

    async def send_bootloader_packet(self, data, **kwds) -> None:
        """Shortcut to ``self.send_packet()`` with the CRTP port and channel
        set to the values that make sure that the bootloader will process the
        command.

        The bootloader responds only to packets sent on the CRTP link control
        port, channel 3.
        """
        return await self.send_packet(
            port=CRTPPort.LINK_CONTROL,
            channel=LinkControlChannel.BOOTLOADER,
            data=data,
            **kwds
        )

    async def validate(self) -> None:
        """Ensures that the information about the flashing targets of the
        bootloader is already retrieved.
        """
        if self._targets is None:
            self._targets = await self._get_targets()

    async def _get_target_info(
        self, target_type: BootloaderTargetType
    ) -> BootloaderTarget:
        response = await self.run_bootloader_command(
            command=(target_type, BootloaderCommand.GET_TARGET_INFO)
        )
        return BootloaderTarget.from_bytes(self, target_type, response)

    async def _get_targets(self) -> List[BootloaderTarget]:
        """Loads information about the possible bootloader targets from the
        bootloader.
        """
        result = [await self._get_target_info(BootloaderTargetType.STM32)]

        if result[-1].protocol_version == BootloaderProtocolVersion.CF2:
            # On the CF2 we also have an NRF32 target
            result.append(await self._get_target_info(BootloaderTargetType.NRF51))

        return result


async def test():
    from aiocflib.crazyflie import Crazyflie
    from tqdm import tqdm

    uri = "radio://0/80/2M/E7E7E7E704"
    async with Crazyflie(uri) as cf:
        uri = await cf.reboot_to_bootloader()

    async with Bootloader(uri) as bl:
        target = await bl.find_target("stm32")
        firmware = open("cf2.bin", "rb").read()

        with tqdm(desc="Flashing", total=len(firmware), unit=" bytes") as progress:
            await target.write_firmware("cf2.bin", on_progress=progress)

        await bl.reboot_to_firmware()


if __name__ == "__main__":
    from aiocflib.crtp import init_drivers
    from anyio import run

    init_drivers()
    run(test)



from __future__ import annotations

from binascii import hexlify, unhexlify
from collections.abc import Sequence
from enum import IntEnum
from typing import ClassVar, Dict, Union

__all__ = (
    "AddressSpace",
    "BootloaderAddressSpace",
    "CrazyradioAddress",
    "CrazyradioAddressLike",
    "DEFAULT_RADIO_ADDRESS",
    "DEFAULT_RADIO_BROADCAST_ADDRESS",
    "RadioAddressSpace",
    "USBAddressSpace",
    "parse_radio_uri",
)


#: Type alias for Crazyradio addresses
CrazyradioAddress = bytes

#: Type alias for objects that can be converted into Crazyradio addresses
CrazyradioAddressLike = Union[int, bytes, str]

#: The default Crazyradio address
DEFAULT_RADIO_ADDRESS = b"\xe7\xe7\xe7\xe7\xe7"  # type: CrazyradioAddress

#: The default Crazyradio broadcast address
DEFAULT_RADIO_BROADCAST_ADDRESS = b"\xff\xe7\xe7\xe7\xe7"  # type: CrazyradioAddress


class CrazyradioDataRate(IntEnum):
    """Enum representing the data rates supported by the radio."""

    DR_250KPS = 0
    DR_1MPS = 1
    DR_2MPS = 2

    @classmethod
    def from_string(cls, value):
        if isinstance(value, cls):
            return value
        elif isinstance(value, int):
            return cls(value)
        else:
            value = value.upper()
            if value in ("250K", "250KPS", "250KBPS"):
                return cls.DR_250KPS
            elif value in ("1M", "1MPS", "1MBPS"):
                return cls.DR_1MPS
            elif value in ("2M", "2MPS", "2MBPS"):
                return cls.DR_2MPS
            else:
                return cls(value)

    def __str__(self):
        if self is CrazyradioDataRate.DR_2MPS:
            return "2M"
        elif self is CrazyradioDataRate.DR_1MPS:
            return "1M"
        else:
            return "250K"


def parse_radio_uri(
    uri: str,
    allow_prefix: bool = False,
    default_address: CrazyradioAddress = DEFAULT_RADIO_ADDRESS,
) -> Dict:
    """Parses a Crazyflie radio URI and returns a dictionary containing the
    parsed parts.

    The dictionary will have the following keys:

    * ``scheme`` (the URI scheme)

    * ``index`` (the index of the radio if multiple radio devices are present)

    * ``channel`` (the channel index)

    * ``data_rate`` (the data rate)

    * ``address`` (the parsed address)

    Parameters:
        uri: the URI to parse
        allow_prefix: whether to allow the user to specify only a prefix of the
            Crazyradio address (where we will assume that the remaining bytes
            are all zeros)
        default_address: the default address to use when the address is omitted
            from the URI
    """
    scheme, sep, path = uri.partition("://")
    if not sep:
        raise ValueError("URI must have a scheme")

    if not path:
        raise ValueError("path must not be empty")

    if path[0] == "/":
        path = path[1:]

    if not path or path == "/":
        path = []
    else:
        path = path.split("/")
        if path[0] == "":
            path.pop(0)

    # Parse index
    if path:
        index = path.pop(0)
        try:
            index = int(index)
        except ValueError:
            raise ValueError("Invalid radio index: {0!r}".format(index)) from None
    else:
        index = 0

    # Parse channel
    if path:
        channel = path.pop(0)
        try:
            channel = int(channel)
        except ValueError:
            raise ValueError("Invalid channel index: {0!r}".format(channel)) from None
        if channel < 0 or channel > 125:
            raise ValueError("Invalid channel index: {0!r}".format(channel))
    else:
        channel = 2

    # Parse data rate
    if path:
        data_rate = path.pop(0)
        try:
            data_rate = CrazyradioDataRate.from_string(data_rate)
        except ValueError:
            raise ValueError("Invalid data rate: {0!r}".format(data_rate)) from None
    else:
        data_rate = CrazyradioDataRate.DR_2MPS

    # Parse address
    if path:
        address = path.pop(0)
        try:
            address = to_radio_address(address, allow_prefix=allow_prefix)
        except Exception as ex:
            raise ValueError("Invalid address: {0!r}".format(address)) from ex
    else:
        address = default_address

    # Extra parts at the end
    if path:
        raise ValueError("Excess parts at the end of the path")

    return {
        "address": address,
        "channel": channel,
        "data_rate": data_rate,
        "index": index,
        "scheme": scheme,
    }


def to_radio_address(
    address: CrazyradioAddressLike, *, allow_prefix: bool = False
) -> CrazyradioAddress:
    """Converts a Crazyradio address-like object to a valid address.

    When the input is a bytes object of length 5, it is returned intact.

    When the input is an integer between 0 and 255, inclusive, it is
    appended in hexadecimal form to E7E7E7E7 and the extended byte sequence
    is returned.

    When the input is a hexadecimal string of length 10, it is unhexlified
    and returned as a bytes object.

    Parameters:
        address: the object to convert into a Crazyradio address
        allow_prefix: whether to allow address prefixes (i.e. addresses that
            have less than five bytes). In such cases, the remaining bytes
            of the address are assumed to be zero.
    """
    if isinstance(address, int) and address >= 0 and address <= 255:
        return RadioAddressSpace.DEFAULT.get_address_for(address)

    if isinstance(address, bytes):
        if len(address) == 5:
            return address
        elif len(address) < 5 and allow_prefix:
            address += bytes((0x00,)) * (5 - len(address))
            return address

    if isinstance(address, str) and len(address) % 2 == 0:
        try:
            address = unhexlify(address)
            if len(address) == 5:
                return address
            elif len(address) < 5 and allow_prefix:
                address += bytes((0x00,)) * (5 - len(address))
                return address
        except Exception:
            pass

    if isinstance(address, str):
        return to_radio_address(int(address))

    raise TypeError(
        "expected a bytes object of length 5, a hexadecimal string of "
        "length 10 or an integer between 0 and 255, inclusive, "
        "got {0!r}".format(address)
    )


class AddressSpace(Sequence):
    """Crazyflie address space that can take integers and return a corresponding
    Crazyflie URI address according to some preconfigured rule.
    """

    pass


class BootloaderAddressSpace(AddressSpace):
    """Crazyflie address space that returns URIs where the Crazyflie bootloader
    could be listening.
    """

    DEFAULT: ClassVar["BootloaderAddressSpace"]

    def __init__(self, index: int = 0, scheme: str = "radio"):
        """Constructor.

        Parameters:
            index: the index of the Crazyradio to return in the URIs of the
                address space
            scheme: the URI address scheme to return in the URIs of the address
                space
        """
        self._index = int(index)
        self._items = [
            "{0}://{1}/{2}".format(scheme, index, channel) for channel in (0, 110)
        ]

    def __getitem__(self, index):
        return self._items[index]

    def __len__(self):
        return len(self._items)


class RadioAddressSpace(AddressSpace):
    """Crazyflie radio address space that returns radio addresses using a
    specified radio index, channel, data rate and address prefix.

    For instance, to assign numbers from 0 to 255 to Crazyflies with
    addresses E7E7E7E700 to E7E7E7E7FF using 2MPS data rate and channel 80:

        addresses = RadioAddressSpace(
            index=0,
            channel=80,
            data_rate="2M",
            prefix="E7E7E7E7",
            length=256
        )

    Alternatively, you can use URI notation:

        addresses = RadioAddressSpace.from_uri("radio://0/80/2M/E7E7E7E7")
    """

    DEFAULT: ClassVar["RadioAddressSpace"]

    @classmethod
    def from_uri(cls, uri: str, length: int = 256):
        """Constructs a RadioAddressSpace from its URI representation.

        Parameters:
            uri: the URI representation of the radio address space
            length: the number of addresses in the constructed address space
        """
        parts = parse_radio_uri(uri, allow_prefix=True)
        parts["prefix"] = parts.pop("address", "E7E7E7E7")
        parts["length"] = int(length)
        return cls(**parts)

    def __init__(
        self,
        index: int = 0,
        channel: int = 80,
        data_rate: Union[CrazyradioDataRate, str] = CrazyradioDataRate.DR_2MPS,
        prefix: Union[bytes, str] = "E7E7E7E7",
        length: int = 256,
        scheme: str = "radio",
    ):
        """Constructor.

        Parameters:
            index: the index of the Crazyradio to return in the URIs of the
                address space
            channel: the channel of the Crazyradio to return in the URIs of the
                address space
            data_rate: the data rate of the Crazyradio to return in the URIs
                of the address space
            prefix: the leading non-zero bytes of the addresses to return in the
                URIs of the address space
            length: the size of the address space; the first address will
                correspond to the address obtained from the prefix
            scheme: the URI address scheme to return in the URIs of the address
                space
        """
        self._index = int(index)
        self._channel = int(channel)
        self._data_rate = CrazyradioDataRate.from_string(data_rate)
        self._scheme = str(scheme)
        self._prefix = to_radio_address(prefix, allow_prefix=True)
        self._length = max(0, int(length))

        self._base_address = int.from_bytes(self._prefix, byteorder="big")
        self._uri_prefix = "{0}://{1}/{2}/{3}".format(
            self._scheme, self._index, self._channel, str(self._data_rate)
        )
        self._uri_format = self._uri_prefix + "/{0}"

    def get_address_for(self, index: int) -> CrazyradioAddress:
        """Returns a Crazyradio address for the drone with the given index in
        this address space.

        This function returns the address only; if you need the full URI, use
        the address space as if it was a Python list.
        """
        if index >= 0 and index < self._length:
            address = self._base_address + index
            return address.to_bytes(5, byteorder="big")
        else:
            raise IndexError

    @property
    def uri_prefix(self) -> str:
        """Returns the URI prefix of this address space, without the section
        reserved for the address.
        """
        return self._uri_prefix

    def __getitem__(self, index: int) -> str:
        address = self.get_address_for(index)
        return self._uri_format.format(hexlify(address).decode("ascii").upper())

    def __len__(self) -> int:
        return self._length


class USBAddressSpace(AddressSpace):
    """Crazyflie address space that returns USB addresses.

    Address zero will belong to the first Crazyflie attached directly to the
    computer; address one will belong to the second Crazyflie and so on.

    This address space has a fixed size of 32 by default. If you want an
    accurate length, you need to construct a USBAddressSpace_ asynchronously
    with the `create()` class method; this will retrieve the number of connected
    Crazyflies at construction time. Alternatively, you may use the `refresh()`
    method of the address space to re-scan the USB bus.
    """

    @classmethod
    async def create(cls):
        result = cls()
        await result.refresh()
        return result

    def __init__(self, length: int = 32, scheme: str = "usb"):
        """Constructor.

        Parameters:
            length: the length of the address space
            scheme: the URI address scheme to return in the URIs of the address
                space
        """
        self._length = max(0, int(length))
        self._format_str = "{0}://".format(scheme) + "{0}"

    def __getitem__(self, index: int) -> str:
        if index >= 0 and index < self._length:
            return self._format_str.format(index)
        else:
            raise IndexError

    def __len__(self) -> int:
        return self._length

    async def refresh(self) -> None:
        """Re-scans the USB bus for connected Crazyflie drones."""
        from aiocflib.drivers.cfusb import CfUsb

        self._length = len(await CfUsb.detect_all())


#: Default bootloader address space for the first connected Crazyradio
BootloaderAddressSpace.DEFAULT = BootloaderAddressSpace()


#: Default radio address space that is a sensible choice for drone swarms with
#: less than 256 drones
RadioAddressSpace.DEFAULT = RadioAddressSpace()



from __future__ import annotations

from anyio import open_file
from enum import IntEnum
from math import ceil
from struct import Struct
from typing import Optional, Union

from aiocflib.utils import chunkify

from .types import BootloaderCommand, BootloaderProtocolVersion, ProgressHandler

__all__ = ("BootloaderTarget", "BootloaderTargetType")


class BootloaderTargetType(IntEnum):
    """Enum representing the CPU targets that the bootloader can flash with a
    new firmware.
    """

    NRF51 = 0xFE
    STM32 = 0xFF

    @property
    def description(self) -> str:
        """Returns a human-readable description of the CPU target."""
        return _target_descriptions.get(self, "Unknown")

    @classmethod
    def from_string(cls, name: Union[str, "BootloaderTargetType"]):
        if isinstance(name, cls):
            return name

        assert isinstance(name, str)
        name = name.lower()

        for key, value in _target_descriptions.items():
            if value.lower() == name:
                return key

        raise ValueError("no such bootloader target: {0!r}".format(name))


_target_descriptions = {
    BootloaderTargetType.NRF51: "nRF51",
    BootloaderTargetType.STM32: "STM32",
}


class BootloaderTarget:
    """Class representing a flashing target for the Crazyflie bootloader.

    This value class is populated from a response to the ``GET_TARGET_INFO``
    command of the bootloader.

    Attributes:
        id: the identifier of the flashing target in the bootloader
        protocol_version: the bootloader protocol version
        page_size: the page size of the flashing process, in bytes
        buffer_pages: number of pages in the flashing buffer
        flash_pages: number of pages in the flash memory where we can write
        start_page: the first page where we can write the firmware
        cpu_id: the CPU ID of the target CPU
    """

    _load_buffer_command_struct = Struct("<BBHH")
    _read_buffer_command_struct = Struct("<BBHH")
    _read_command_struct = Struct("<BBHH")
    _write_command_struct = Struct("<BB")
    _write_command_params_struct = Struct("<HHH")
    _target_struct = Struct("<HHHH12s")

    _LOAD_BUFFER_CHUNK_SIZE = 25
    _READ_FLASH_CHUNK_SIZE = 25

    @classmethod
    def from_bytes(cls, owner, id: BootloaderTargetType, data: bytes):
        """Constructs a BootloaderTarget_ instance from its raw byte-level
        representation in the appropriate CRTP packet.

        Parameters:
            owner: the bootloader instance that owns this target
            id: the bootloader target type that was used when sending the
                CRTP packet whose response we are decoding
            data: the data section of the CRTP response, without the command
                bytes

        Returns:
            an appropriately constructed BootloaderTarget_ instance
        """
        result = cls(owner, id)

        (
            page_size,
            buffer_pages,
            flash_pages,
            start_page,
            cpu_id,
        ) = cls._target_struct.unpack(data[: cls._target_struct.size])

        if len(data) > cls._target_struct.size:
            protocol_version = BootloaderProtocolVersion(data[cls._target_struct.size])
        else:
            protocol_version = BootloaderProtocolVersion.UNKNOWN

        result.page_size = page_size
        result.buffer_pages = buffer_pages
        result.flash_pages = flash_pages
        result.start_page = start_page
        result.cpu_id = cpu_id
        result.protocol_version = protocol_version

        return result

    def __init__(self, owner, id: BootloaderTargetType):
        """Constructor."""
        self._bootloader = owner

        self.id = id
        self.protocol_version = BootloaderProtocolVersion.UNKNOWN
        self.page_size = 0  # type: int
        self.buffer_pages = 0  # type: int
        self.flash_pages = 0  # type: int
        self.start_page = 0  # type: int
        self.cpu_id = b""

    def __str__(self):
        result = [
            "Target info: {0} (0x{1:X})".format(self.id.description, self.id),
            "Flash pages: {0.flash_pages} | Page size: {0.page_size} | "
            "Buffer pages: {0.buffer_pages} | Start page: {0.start_page}".format(self),
            "{0.max_firmware_size_in_kbytes} KBytes of flash available for firmware image.".format(
                self
            ),
        ]
        return "\n".join(result)

    @property
    def buffer_size(self) -> int:
        """Returns the size of the upload buffer on this target, in bytes."""
        return self.buffer_pages * self.page_size

    @property
    def firmware_address(self) -> int:
        """Address where the firmware should be written in the flash memory."""
        return self.start_page * self.page_size

    @property
    def flash_size(self) -> int:
        """Returns the size of the flash memory available for the firmware image
        on this target, in bytes.
        """
        return self.flash_pages * self.page_size

    @property
    def flash_size_in_kbytes(self) -> int:
        """Returns the size of the flash memory available for the firmware image
        on this target, in KBytes.
        """
        return self.flash_size // 1024

    @property
    def max_firmware_size(self) -> int:
        """Returns the size of the flash memory available for the firmware image
        on this target, in bytes.
        """
        return self.flash_size - self.firmware_address

    @property
    def max_firmware_size_in_kbytes(self) -> int:
        """Returns the size of the flash memory available for the firmware image
        on this target, in KBytes.
        """
        return self.max_firmware_size // 1024

    async def read_flash(
        self,
        address: int = 0,
        length: int = -1,
        *,
        on_progress: Optional[ProgressHandler] = None,
    ) -> bytes:
        """Reads the given number of bytes from the given address of the flash
        memory of the target.

        Parameters:
            address: the address to read from
            length: the maximum number of bytes to read. Negative numbers mean
                to read everything until the end of the flash memory.
            on_progress: function to call periodically with the number of bytes
                read in each iteration

        Returns:
            the contents of the flash memory, starting from the given address.
        """
        if address < 0:
            raise ValueError("address cannot be negative")

        result = []
        to_read = length if length >= 0 else (self.flash_size - address)

        # Special support for tqdm progress bars
        if hasattr(on_progress, "reset") and hasattr(on_progress, "update"):
            on_progress.reset(to_read)  # type: ignore
            on_progress = on_progress.update  # type: ignore

        while to_read > 0:
            page, offset = divmod(address, self.page_size)
            data = await self._bootloader.run_bootloader_command(
                command=self._read_command_struct.pack(
                    self.id, BootloaderCommand.READ_FLASH, page, offset
                )
            )
            result.append(data)

            bytes_read = len(data)
            address += bytes_read
            to_read -= bytes_read

            if on_progress:
                on_progress(bytes_read)

            if bytes_read < self._READ_FLASH_CHUNK_SIZE:
                # end of flash
                break

        return b"".join(result)

    async def read_firmware(
        self, length: int = -1, *, on_progress: Optional[ProgressHandler] = None
    ) -> bytes:
        """Reads the given number of bytes from the firmware area of the flash
        memory of the target.

        Parameters:
            length: the maximum number of bytes to read. Negative numbers mean
                to read everything until the end of the flash memory.
            on_progress: function to call periodically with the number of bytes
                read in each iteration

        Returns:
            the contents of the flash memory, starting from the given address.
        """
        return await self.read_flash(
            self.firmware_address, length, on_progress=on_progress
        )

    async def write_flash(
        self,
        address: int,
        data: bytes,
        *,
        on_progress: Optional[ProgressHandler] = None,
    ) -> None:
        """Writes some data at the given address into the flash memory of the
        target.

        Parameters:
            address: the address to write the data to. Right now it must point
                to the start of a page on the flash memory.
            data: the data to write
            on_progress: function to call periodically with the number of bytes
                written during the operation
        """
        if address % self.page_size:
            raise ValueError("write_flash() address must point to the start of a page")

        # Special support for tqdm progress bars
        if hasattr(on_progress, "reset") and hasattr(on_progress, "update"):
            on_progress.reset(len(data))  # type: ignore
            on_progress = on_progress.update  # type: ignore

        for start, size in chunkify(0, len(data), step=self.buffer_size):
            await self._fill_buffer_with(
                data[start : (start + size)], on_progress=on_progress
            )
            await self._flush_buffer_to_flash(address, size)

            address += size

    async def write_firmware(
        self,
        firmware: Union[bytes, str],
        *,
        on_progress: Optional[ProgressHandler] = None,
    ) -> None:
        """Writes the given data to the firmware area of the flash memory of
        the target.

        Parameters:
            firmware: the firmware to write. When it is a string, it is treated
                as the name of a file containing the firmware. When it is a
                bytes object, it is treated as the firmware itself.
            on_progress: function to call periodically with the number of bytes
                written during the operation
        """
        if isinstance(firmware, str):
            async with await open_file(firmware, "rb") as fp:
                firmware = await fp.read()

        assert isinstance(firmware, bytes)

        await self.write_flash(self.firmware_address, firmware, on_progress=on_progress)

    async def _fill_buffer_with(
        self,
        data: bytes,
        *,
        validate: bool = False,
        on_progress: Optional[ProgressHandler] = None,
    ) -> None:
        """Fills the upload buffer on the target with the given data.

        Parameters:
            data: the data to write into the buffer
        """
        assert len(data) <= self.buffer_size

        # First we fill the entire buffer, then, if we need to validate the
        # result, we validate in one batch after uploading. This is to ensure
        # that the Crazyflie has time to process the inbound packets before we
        # start reading the buffer back.
        for start, size in chunkify(0, len(data), step=self._LOAD_BUFFER_CHUNK_SIZE):
            page, offset = divmod(start, self.page_size)
            to_write = data[start : (start + size)]
            await self._bootloader.send_bootloader_packet(
                self._load_buffer_command_struct.pack(
                    self.id, BootloaderCommand.LOAD_BUFFER, page, offset
                )
                + to_write
            )

            if on_progress:
                on_progress(size)

        # Now we validate if needed
        if validate:
            errors = []
            for index, (start, size) in enumerate(
                chunkify(0, len(data), step=self._LOAD_BUFFER_CHUNK_SIZE)
            ):
                page, offset = divmod(start, self.page_size)
                expected = data[start : (start + size)]

                observed = await self._bootloader.run_bootloader_command(
                    command=self._read_buffer_command_struct.pack(
                        self.id, BootloaderCommand.READ_BUFFER, page, offset
                    ),
                    timeout=0.1,
                )

                if observed[:size] != expected:
                    from hexdump import hexdump

                    print("Tried to upload:")
                    hexdump(expected)
                    print()
                    print("Currently in buffer:")
                    hexdump(observed[:size])
                    print()

                    errors.append(index)

            if errors:
                print(repr(errors))
                raise IOError("failed to update buffer")

    async def _flush_buffer_to_flash(self, start: int, size: int) -> None:
        start, remainder = divmod(start, self.page_size)
        assert remainder == 0

        num_pages = int(ceil(size / self.page_size))

        # Note that we use a timeout of 2.5 seconds here and we don't re-send
        # this packet. This is intentional; sometimes the flash request takes
        # more than one second, and the STM32 bootloader has a problem with
        # re-sent flash requests.
        result = await self._bootloader.run_bootloader_command(
            command=self._write_command_struct.pack(
                self.id, BootloaderCommand.WRITE_FLASH
            ),
            data=self._write_command_params_struct.pack(0, start, num_pages),
            timeout=2.5,
            attempts=3,
        )

        if len(result) < 2:
            raise IOError("invalid response from flash write request")

        done = result[0] > 0
        status = result[1]

        if status == 1:
            raise IOError("invalid write request sent to target")
        elif status == 2:
            raise IOError("failed to erase sector in flash memory")
        elif status == 3:
            raise IOError("failed to write new data into flash memory")
        elif status > 0:
            raise IOError("unknown error (code = {0})".format(status))
        elif not done:
            raise IOError("target says write is not done but returned no error code")



"""Various utilities that is needed by the asynchronous Crazyflie library."""

from .chunks import chunkify
from .timing import timing

__all__ = ("anop", "chunkify", "nop", "timing")


def nop(*args, **kwds) -> None:
    """Dummy function that can be invoked with arbitrary arguments and that
    does nothing.
    """
    pass


async def anop(*args, **kwds) -> None:
    """Dummy async function that can be invoked with arbitrary arguments and that
    does nothing.
    """
    pass



from .base import CRTPDriver
from .registry import register

__all__ = ("CRTPDriver", "init_drivers", "register")


def init_drivers():
    """Initializes all the commonly supported and used drivers and middleware
    in the CRTP stack.

    This method is provided as a convenience so you don't need to import the
    drivers and middleware you wish to use in the driver URIs before
    constructing a driver object.
    """
    from .cpplink import CppRadioDriver  # noqa
    from .radio import RadioDriver  # noqa
    from .sitl import SITLDriver  # noqa
    from .usb import USBDriver  # noqa

    from ..middleware.log import LoggingMiddleware  # noqa



from contextlib import asynccontextmanager
from typing import Any, List, Optional
from urllib.parse import urlparse

from aiocflib.crtp.crtpstack import CRTPPacket
from aiocflib.drivers.cfusb import CfUsb
from aiocflib.utils.concurrency import ObservableValue

from ..exceptions import WrongURIType

from .base import CRTPDriver
from .registry import register


_link_quality = None  # type: Optional[ObservableValue[float]]


@register("usb")
class USBDriver(CRTPDriver):
    """CRTP driver that allows us to communicate with a Crazyflie that is
    connected directly via a USB cable.
    """

    _device: Any

    @asynccontextmanager
    async def _connected_to(self, uri: str):
        parts = urlparse(uri)

        try:
            index = int(parts.netloc)
        except ValueError:
            raise WrongURIType("Invalid USB URI: {0!r}".format(uri)) from None

        if index < 0:
            raise WrongURIType("USB port index must be non-negative")

        device = await CfUsb.detect_one(index=index)
        device.use_crtp_to_usb = True

        try:
            async with device as self._device:
                yield self
        finally:
            self._device = None

    @property
    def index(self) -> Optional[int]:
        """The index of the USB device on the USB hub, or `None` if not known."""
        if not self.uri:
            return None

        parts = urlparse(self.uri)
        try:
            return int(parts.netloc)
        except ValueError:
            return None

    @property
    def is_safe(self) -> bool:
        return True

    @property
    def link_quality(self) -> ObservableValue[float]:
        global _link_quality
        if _link_quality is None:
            _link_quality = ObservableValue.constant(1.0)
        return _link_quality

    @property
    def name(self) -> str:
        return "USBCDC"

    async def receive_packet(self) -> CRTPPacket:
        """Receives a single CRTP packet.

        Returns:
            the next CRTP packet that was received
        """
        data = await self._device.receive_bytes()
        return CRTPPacket.from_bytes(data)

    async def send_packet(self, packet: CRTPPacket):
        """Sends a CRTP packet.

        Parameters:
            packet: the packet to send
        """
        await self._device.send_bytes(packet.to_bytes())

    @classmethod
    async def scan_interfaces(cls) -> List[str]:
        """Scans all interfaces of this type for available Crazyflie quadcopters
        and returns a list with appropriate connection URIs that could be used
        to connect to them.

        Returns:
            the list of connection URIs where a Crazyflie was detected; an empty
            list is returned for interfaces that do not support scanning
        """
        devices = await CfUsb.detect_all()
        return ["usb://{0}" for index in range(len(devices))]



"""Base class for middlewares that forward each call to an underlying CRTP
driver instance before / after performing some operations on the call
arguments.
"""

from contextlib import asynccontextmanager
from typing import Optional

from aiocflib.crtp.crtpstack import CRTPPacket
from aiocflib.crtp.drivers.base import CRTPDriver
from aiocflib.utils.concurrency import ObservableValue


class MiddlewareBase(CRTPDriver):
    """Base class for middlewares that forward each call to an underlying CRTP
    driver instance before / after performing some operations on the call
    arguments.
    """

    def __init__(self, wrapped: CRTPDriver):
        """Constructor.

        Parameters:
            wrapped: the CRTP driver that the middleware wraps
        """
        self._wrapped = wrapped
        self._init()

    @asynccontextmanager
    async def _connected_to(self, uri: str):
        async with self._wrapped._connected_to(uri):
            yield

    def _init(self) -> None:
        pass

    @property
    def is_safe(self) -> bool:
        return self._wrapped.is_safe

    @property
    def link_quality(self) -> ObservableValue[float]:
        return self._wrapped.link_quality

    @property
    def name(self) -> str:
        return self._wrapped.name

    @property
    def uri(self) -> Optional[str]:
        return self._wrapped.uri

    @property
    def wrapped(self) -> CRTPDriver:
        """Returns the CRTP driver wrapped by the middleware."""
        return self._wrapped

    async def notify_rebooted(self) -> None:
        return await self._wrapped.notify_rebooted()

    async def receive_packet(self) -> CRTPPacket:
        return await self._wrapped.receive_packet()

    async def send_packet(self, packet: CRTPPacket):
        return await self._wrapped.send_packet(packet)

    async def use_safe_link(self):
        return await self._wrapped.use_safe_link()



"""Interface specificaton for asynchronous CRTP driver implementations."""

from abc import ABCMeta, abstractmethod, abstractproperty
from contextlib import asynccontextmanager
from typing import AsyncContextManager, List, Optional

from aiocflib.crtp.crtpstack import CRTPPacket
from aiocflib.utils.concurrency import ObservableValue

from ..exceptions import WrongURIType

__author__ = "CollMot Robotics Ltd"
__all__ = ("CRTPDriver",)


class CRTPDriver(metaclass=ABCMeta):
    """Interface specificaton for asynchronous CRTP driver implementations.

    All asynchronous CRTP driver implementations must inherit from this class.
    """

    @staticmethod
    @asynccontextmanager
    async def connected_to(uri: str, *args, **kwds):
        """Creates a CRTPDriver_ instance from a URI specification and connects
        to the Crazyflie located at the given URI. Closes the connection when the
        execution exits the context.

        Parameters:
            uri: the URI to create the driver instance from

        Additional positional and keyword arguments are forwarded to the
        driver factory determined from the URI scheme.
        """
        from .registry import find as find_driver
        from ..middleware.registry import find as find_middleware

        scheme, sep, rest = uri.partition("://")
        if not sep:
            raise ValueError("CRTP driver URI must contain ://")

        scheme, *middleware_names = scheme.split("+")

        try:
            driver_factory = find_driver(scheme)
        except KeyError:
            raise WrongURIType(
                "Unknown CRTP driver URI: {0!r}".format(scheme)
            ) from None

        driver = driver_factory(*args, **kwds)
        driver.uri = uri

        for middleware_name in middleware_names:
            try:
                middleware = find_middleware(middleware_name)
            except KeyError:
                raise WrongURIType(
                    "Unknown middleware in URI: {0!r}".format(middleware_name)
                ) from None
            driver = middleware(driver)

        async with driver._connected_to(uri):
            yield driver

    @abstractmethod
    def _connected_to(self, uri: str) -> AsyncContextManager[None]:
        """Connects the driver instance to a specified URI.

        This method is not public; use the `connected_to()` async context manager
        instead, which ensures that the link is closed when the execution leaves
        the context.

        Since the method is not meant for public use, it does _not_ check whether
        the received URI has a scheme that corresponds to the driver; it is
        assumed that the caller already took care of it.

        Parameters:
            uri: the URI to connect the driver to

        Returns:
            an async context manager that connects to the specified URI when
            the context is entered and that closes the connection when the
            context is exited
        """
        raise NotImplementedError

    @abstractproperty
    def is_safe(self) -> bool:
        """Returns whether this link is safe.

        A safe link guarantees that each packet that was sent via the link
        eventually gets delivered to the peer, _or_ that the link gets closed
        if the delivery fails.
        """
        raise NotImplementedError

    @abstractproperty
    def link_quality(self) -> ObservableValue[float]:
        """Observable measurement of link quality, as a float between 0 (no
        link) and 1 (perfect link).
        """
        raise NotImplementedError

    @abstractproperty
    def name(self) -> str:
        """Returns a human-readable name of the interface."""
        raise NotImplementedError

    async def notify_rebooted(self) -> None:  # noqa: B027
        """Notifies the driver that the underlying Crazyflie device has been
        rebooted. The drivers may respond to this request by scheduling some
        setup operations to perform on the link again in an attempt to restore
        the link to the state before the reboot.
        """
        pass

    @abstractmethod
    async def receive_packet(self) -> CRTPPacket:
        """Receives a single CRTP packet.

        Returns:
            the next CRTP packet that was received
        """
        raise NotImplementedError

    @classmethod
    async def scan_interfaces(cls) -> List[str]:
        """Scans all interfaces of this type for available Crazyflie quadcopters
        and returns a list with appropriate connection URIs that could be used
        to connect to them.

        Returns:
            the list of connection URIs where a Crazyflie was detected; an empty
            list is returned for interfaces that do not support scanning
        """
        return []

    @abstractmethod
    async def send_packet(self, packet: CRTPPacket) -> None:
        """Sends a CRTP packet.

        Parameters:
            packet: the packet to send
        """
        raise NotImplementedError

    @property
    def uri(self) -> Optional[str]:
        """The URI with which the driver was created, if known."""
        return getattr(self, "_uri", None)

    @uri.setter
    def uri(self, value: str) -> None:
        self._uri = value

    async def use_safe_link(self) -> None:  # noqa: B027
        """Notifies the driver that the caller wishes to use a safe link where
        it does not have to worry about packet loss.

        When the driver supports safe link mode, it should switch into a mode
        where this can be ensured.
        """
        pass



"""Classes related to sending motor-related commands to a Crazyflie."""

from anyio import sleep
from typing import Iterable, Optional

from .crazyflie import Crazyflie

__all__ = ("Motors",)


class Motors:
    """Class representing the motors of a Crazyflie instance."""

    _crazyflie: Crazyflie

    def __init__(self, crazyflie: Crazyflie):
        """Constructor.

        Parameters:
            crazyflie: the Crazyflie for which we need to handle motor-related
                messages
        """
        self._crazyflie = crazyflie

    async def test(
        self,
        indices: Optional[Iterable[int]] = None,
        power: int = 20000,
        duration: float = 1,
        delay: float = 1,
    ) -> None:
        """Tests the motors by spinning them up and down.

        Parameters:
            indices: the 1-based indices of the motors to test; `None` means all
                the motors
            power: the power value to send to the motors; must be an unsigned
                integer in the range [0; 65535]. The default will spin up the
                motor but it will not spin fast enough to cause any harm on a
                standard Crazyflie.
            duration: the duration of the motor test, in seconds
            delay: the delay between consecutive tests, in seconds
        """
        cf = self._crazyflie
        if indices is None:
            indices = (1, 2, 3, 4)

        async with cf.parameters.set_and_restore("motorPowerSet.enable", 1, 0):
            for index, _motor_index in enumerate(indices):
                param_name = f"motorPowerSet.m{index + 1}"
                if index > 0:
                    await sleep(delay)
                await cf.parameters.set(param_name, power)
                await sleep(duration)
                await cf.parameters.set(param_name, 0)



"""asyncio-based variant of the Crazyflie Python library."""

from .errors import error_to_string
from .version import __version__, __version_info__

__all__ = ("error_to_string", "__version__", "__version_info__")



"""Enumeration types related to the bootloader of the Crazyflie."""

from enum import IntEnum
from typing import Callable

__all__ = ("BootloaderProtocolVersion",)


class BootloaderCommand(IntEnum):
    """Enum describing the commands that the Crazyflie or its bootloader
    handles on the LINK_CONTROL CRTP port, channel 3 (bootloader).

    Commands 0x10 - 0x1F are responded to only when the Crazyflie is in
    bootloader mode. All other commands are responded to only when the
    Crazyflie is in firmware mode.
    """

    SHUTDOWN = 0x01
    SUSPEND = 0x02
    RESUME = 0x03
    GET_BATTERY_VOLTAGE = 0x04

    GET_TARGET_INFO = 0x10
    SET_ADDRESS = 0x11
    GET_MAPPING = 0x12
    LOAD_BUFFER = 0x14
    READ_BUFFER = 0x15
    WRITE_FLASH = 0x18
    READ_FLASH = 0x1C

    RESET = 0xF0
    RESET_INIT = 0xFF


class BootloaderProtocolVersion(IntEnum):
    CF1_V0 = 0x00
    CF1_V1 = 0x01
    CF2 = 0x10
    UNKNOWN = 0xFF

    @property
    def description(self) -> str:
        """Returns a human-readable description of the bootloader protocol
        version.
        """
        return _bootloader_protocol_descriptions.get(self, "Unknown")

    @property
    def is_cf2(self) -> bool:
        """Returns whether this protocol version corresponds to a Crazyflie
        2.0 instance.
        """
        return self is BootloaderProtocolVersion.CF2


_bootloader_protocol_descriptions = {
    BootloaderProtocolVersion.CF1_V0: "Crazyflie Nano Quadcopter (1.0)",
    BootloaderProtocolVersion.CF1_V1: "Crazyflie Nano Quadcopter (1.0)",
    BootloaderProtocolVersion.CF2: "Crazyflie 2.0",
}


#: Type alias for progress handler functions in BootloaderTarget_
ProgressHandler = Callable[[float], None]



"""Classes for modeling and handling CRTP packets."""

from anyio import create_memory_object_stream
from anyio.abc import ObjectReceiveStream
from array import array
from collections import defaultdict
from contextlib import asynccontextmanager, contextmanager
from enum import IntEnum
from functools import partial
from inspect import iscoroutinefunction
from typing import (
    AsyncIterator,
    Awaitable,
    Callable,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Sequence,
    Union,
)

from aiocflib.utils.concurrency import AwaitableValue

__author__ = "CollMot Robotics Ltd"
__all__ = (
    "CRTPCommandLike",
    "CRTPDataLike",
    "CRTPDispatcher",
    "CRTPPacket",
    "CRTPPort",
    "CRTPPortLike",
    "LinkControlChannel",
)


#: Mapping from CRTP port names to short three-letter identifiers
_crtp_port_codes: List[str] = [
    "CON",
    "P01",
    "PRM",
    "CMD",
    "MEM",
    "LOG",
    "LOC",
    "GEN",
    "HLC",
    "P09",
    "P10",
    "P11",
    "P12",
    "PLT",
    "DBG",
    "LNK",
]


class CRTPPort(IntEnum):
    """Enum representing the available ports of the CRTP protocol."""

    CONSOLE = 0x00
    UNUSED_1 = 0x01
    PARAMETERS = 0x02
    COMMANDER = 0x03
    MEMORY = 0x04
    LOGGING = 0x05
    LOCALIZATION = 0x06
    GENERIC_COMMANDER = 0x07
    HIGH_LEVEL_COMMANDER = 0x08
    UNUSED_9 = 0x09
    UNUSED_10 = 0x0A
    UNUSED_11 = 0x0B
    UNUSED_12 = 0x0C
    PLATFORM = 0x0D
    DEBUG = 0x0E
    LINK_CONTROL = 0x0F

    @property
    def code(self) -> str:
        return _crtp_port_codes[int(self)]


class LinkControlChannel(IntEnum):
    """Enum representing the names of the link control channels in the link
    control service of the CRTP protocol.

    This is declared here instead of in a dedicated "link control" module
    because it is used in multiple places in the library (especially the
    bootloader channel).
    """

    ECHO = 0
    SOURCE = 1
    SINK = 2
    BOOTLOADER = 3


#: Type alias for objects that can be convered into a CRTP command byte
CRTPCommandLike = Union[int, bytes, Iterable[Union[int, bytes]]]

#: Type alias for objects that can be converted into the data of a CRTP packet
CRTPDataLike = Union[bytes, Sequence[int]]

#: Type alias for objects that can be converted into a CRTP port
CRTPPortLike = Union[int, CRTPPort]


class MemoryType(IntEnum):
    """Enum representing the types of memories supported by a Crazyflie."""

    I2C = 0
    ONE_WIRE = 1
    LED = 0x10
    LOCO = 0x11
    TRAJECTORY = 0x12
    LOCO2 = 0x13
    LIGHTHOUSE = 0x14
    TESTER = 0x15
    SD_CARD = 0x16
    LED_SEQUENCE = 0x17
    APP = 0x18
    DECK = 0x19

    @property
    def description(self) -> str:
        """Human-readable description of the memory type."""
        return _memory_type_descriptions.get(int(self), "Unknown")


_memory_type_descriptions: Dict[int, str] = {
    MemoryType.I2C: "I2C",
    MemoryType.ONE_WIRE: "1-wire",
    MemoryType.LED: "LED driver",
    MemoryType.LOCO: "Loco positioning",
    MemoryType.TRAJECTORY: "Trajectory",
    MemoryType.LOCO2: "Loco positioning 2",
    MemoryType.LIGHTHOUSE: "Lighthouse positioning",
    MemoryType.TESTER: "Memory tester",
    MemoryType.SD_CARD: "SD card",
    MemoryType.LED_SEQUENCE: "LED sequence",
    MemoryType.APP: "Application",
    MemoryType.DECK: "Deck",
}


class CRTPPacket:
    """A single packet that can be sent or received via a CRTP connection."""

    @classmethod
    def from_bytes(cls, data: CRTPDataLike):
        """Constructs a CRTP packet from its raw representation that includes
        the header _and_ the data itself.
        """
        if data is None:
            raise ValueError("data may not be empty")
        if isinstance(data, bytes):
            header = int(data[0])
        else:
            header = data[0]
        return cls(header=header, data=data[1:])

    @classmethod
    def null(cls):
        """Constructs a null CRTP packet."""
        return cls(header=0xFF)

    @classmethod
    def safe_link(cls, value: bool = True):
        """Constructs a special CRTP packet that turns safe link mode on or off.

        Parameters:
            value: whether the safe link mode should be enabled or disabled
        """
        return cls(header=0xFF, data=b"\x05\x01" if value else b"\x05\x00")

    def __init__(
        self,
        header: Optional[int] = None,
        data: Optional[CRTPDataLike] = None,
        *,
        port: Optional[CRTPPortLike] = None,
        channel: int = 0
    ):
        """Constructor.

        Parameters:
            header: the header of the packet. Takes precedence over `port` and
                `channel`.
            data: the data object of the packet
            port: when it is not `None`, and `header` is `None`, specifies the
                CRTP port of the packet
            channel: when `port` is not `None`, and `header` is `None`,
                specifies the CRTP channel of the packet
        """
        self._data = bytes()
        self._header = 0

        if header is not None:
            self.header = header
        elif port is not None:
            self.port = port
            self.channel = channel

        if data is not None:
            self.data = data

    @property
    def channel(self):
        """The channel index of the packet."""
        return self._header & 0x03

    @channel.setter
    def channel(self, value):
        self._header = (self._header & 0xFC) | (value & 0x03)

    @property
    def command(self) -> Optional[int]:
        """Returns the first byte of the data block of the CRTP packet. This
        byte is sometimes used as a command byte in some CRTP packets.
        """
        return self._data[0] if self._data else None

    @property
    def data(self) -> bytes:
        """Get the packet data as a raw immutable bytes object."""
        return self._data

    @data.setter
    def data(self, value) -> None:
        """Set the packet data"""
        if isinstance(value, bytes):
            self._data = value
        elif isinstance(value, str):
            self._data = value.encode("ISO-8859-1")
        elif isinstance(value, (array, bytearray, list, tuple)):
            self._data = bytes(value)
        elif value is None:
            self._data = bytes()
        else:
            raise TypeError(
                "Data must be None, bytes, bytearray, string, list or tuple,"
                " not {}".format(type(value))
            )

    @property
    def header(self) -> int:
        """Returns the header byte of the CRTP packet."""
        return self._header

    @header.setter
    def header(self, value: int):
        # The two bits in position 3 and 4 needs to be set for legacy
        # support of the bootloader, according to the official cflib library
        self._header = (value & 0xFF) | (0x3 << 2)

    @property
    def is_null(self) -> bool:
        """Returns whether this packet is a null packet."""
        return (self._header & 0xF3) == 0xF3 and not self._data

    @property
    def port(self) -> CRTPPort:
        """The CRTP port of the packet."""
        return CRTPPort((self._header & 0xF0) >> 4)

    @port.setter
    def port(self, value) -> None:
        value = int(value)
        self._header = (self._header & 0x0F) | ((value << 4) & 0xF0)

    def to_bytes(self, safelink_bits: int = 12) -> bytes:
        """Convers the packet to its raw byte-level representation."""
        return bytes(((self._header & 0xF3) | safelink_bits,)) + self._data

    __bytes__ = to_bytes

    def __repr__(self):
        """Returns a unique, machine-parseable representation of the packet."""
        return "{0.__class__.__name__}(header={0.header!r}, data={0.data!r})".format(
            self
        )

    def __str__(self):
        """Returns a human-readable string representation of the packet."""
        return "{0}:{1} {2}".format(self.port, self.channel, tuple(self.data))


#: Type alias for functions that can handle CRTP packets
CRTPPacketHandler = Union[
    Callable[[CRTPPacket], None], Callable[[CRTPPacket], Awaitable[None]]
]


class CRTPDispatcher:
    """Auxiliary data structure that allows one to associate queues to
    combinations of CRTP ports and channels, and to dispatch a CRTP packet to
    interested queues.
    """

    def __init__(self):
        """Constructor.

        Creates an empty dispatch table.
        """
        self._by_port_sync = defaultdict(list)
        self._by_port_async = defaultdict(list)

    @asynccontextmanager
    async def create_packet_queue(
        self, port: Optional[CRTPPortLike] = None, *, queue_size: int = 0
    ) -> AsyncIterator[ObjectReceiveStream]:
        """Context manager that creates a queue that will yield incoming
        CRTP packets coming from the given port.

        Parameters:
            port: the CRTP port to match; `None` means to match all CRTP ports
            queue_size: number of pending packets that may stay in the queue
                before blocking. Typically 0 is enough.

        Returns:
            a readable stream in which the dispatcher will place all packets
            that match the given port (or all ports if no port is specified)
        """
        tx_queue, rx_queue = create_memory_object_stream[CRTPPacket](queue_size)
        async with tx_queue:
            with self.registered(tx_queue.send, port=port):
                yield rx_queue

    async def dispatch(self, packet: CRTPPacket) -> None:
        """Handles an incoming CRTP packet and dispatches them to the
        appropriate listener functions.

        Right now this function may block the current task if an asynchronous
        listener is registered on the packet and the listener blocks. If you
        plan to do long-running operations in the listener, you should spawn a
        new task within the listener instead.
        """
        # Send the packet to the port-specific sync and async handlers
        for handler in self._by_port_sync[packet.port]:
            handler(packet)
        for handler in self._by_port_async[packet.port]:
            await handler(packet)

        # Send the packet to the "all packets" handlers
        for handler in self._by_port_sync[None]:
            handler(packet)
        for handler in self._by_port_async[None]:
            await handler(packet)

    def register(
        self, handler: CRTPPacketHandler, *, port: Optional[CRTPPortLike] = None
    ) -> Callable[[], None]:
        """Registers a handler to call for each incoming CRTP packet matching a
        given port.

        Parameters:
            handler: the handler to call. This may be a synchronous or an
                asynchronous function.
            port: the CRTP port that the incoming packet must have in order to
                match the handler. `None` means that any port will match.

        Returns:
            a disposer function that can be called when the handler must be
            deregistered
        """
        if port is not None:
            port = CRTPPort(port)
        if iscoroutinefunction(handler):
            handlers = self._by_port_async[port]
        else:
            handlers = self._by_port_sync[port]
        handlers.append(handler)
        return partial(handlers.remove, handler)

    @contextmanager
    def registered(
        self, handler: CRTPPacketHandler, *, port: Optional[CRTPPortLike] = None
    ):
        """Context manager that registers a given packet handler for a CRTP
        port when the context is entered and deregisters the handler when the
        context is exited.
        """
        disposer = self.register(handler, port=port)
        try:
            yield
        finally:
            disposer()

    @contextmanager
    def wait_for_next_packet(
        self,
        predicate: Optional[Callable[[CRTPPacket], bool]],
        *,
        port: Optional[CRTPPortLike] = None
    ) -> Iterator[AwaitableValue]:
        """Context manager that waits for a CRTP packet on the given port (or
        all ports) matching a given predicate while the execution is in the
        context.

        Returns:
            an awaitable value that resolves to the first CRTP packet matching
            the port number and the predicate
        """
        value: AwaitableValue[CRTPPacket] = AwaitableValue()

        if predicate:

            def matcher(packet: CRTPPacket) -> None:
                if predicate(packet):  # type: ignore
                    value.set(packet)

        else:
            matcher = value.set  # type: ignore

        with self.registered(matcher, port=port):
            yield value



"""Classes related to accessing the memory subsystem of a Crazyflie."""

from abc import abstractmethod, abstractproperty, ABCMeta
from dataclasses import dataclass
from enum import IntEnum
from errno import ENODATA
from struct import Struct, error as StructError
from typing import Callable, ClassVar, List, Optional, Tuple

from aiocflib.crtp import CRTPPort, MemoryType
from aiocflib.errors import error_to_string
from aiocflib.utils import chunkify
from aiocflib.utils.checksum import crc32
from aiocflib.utils.registry import Registry

from .crazyflie import Crazyflie

__all__ = ("Memory", "MemoryType")


class MemoryChannel(IntEnum):
    """Enum representing the names of the channels of the memory service in
    the CRTP protocol.
    """

    INFO = 0
    READ = 1
    WRITE = 2


class MemoryInfoCommand(IntEnum):
    """Enum representing the names of the information commands in the memory
    service of the CRTP protocol.
    """

    GET_NUMBER_OF_MEMORIES = 1
    GET_DETAILS = 2


@dataclass(frozen=True)
class MemoryElement:
    """Class containing information about a single memory element on a
    Crazyflie.
    """

    index: int
    """Index of the memory element."""

    type: int
    """Type identifier of the memory element."""

    size: int
    """Size of the memory element, in bytes."""

    address: int
    """Offset of the memory element in its address space."""

    _struct: ClassVar[Struct] = Struct("<BIQ")

    @classmethod
    def from_bytes(cls, index: int, data: bytes):
        """Constructs a MemoryElement_ instance from its representation in
        the CRTP memory details packet.

        Parameters:
            index: the index of the memory element that is being constructed
            data: the data section of the CRTP packet, without the command byte
                and the ID of the memory element

        Raises:
            ValueError: if the data section cannot be parsed
        """
        try:
            return cls(index, *cls._struct.unpack(data))
        except StructError:
            raise ValueError("invalid memory description") from None


_memory_handler_registry: Registry[
    Callable[[MemoryElement, Crazyflie], "MemoryHandler"]
] = Registry()


class MemoryHandler(metaclass=ABCMeta):
    """Interface specification for memory handlers that know how to read and
    write a certain type of memory.
    """

    #: Maximum number of bytes that can be read in a single request
    MAX_READ_REQUEST_LENGTH = 20

    #: Maximum number of bytes that can be written in a single request
    MAX_WRITE_REQUEST_LENGTH = 25

    @staticmethod
    def for_element(element: MemoryElement, owner: Crazyflie) -> "MemoryHandler":
        """Constructs an appropriate memory handler for the given memory
        element, depending on its type.

        Parameters:
            element: the memory element
            owner: the Crazyflie that owns the memory element
        """
        key = str(element.type)
        cls = _memory_handler_registry.find(key, default=MemoryHandlerBase)
        return cls(element, owner=owner)  # type: ignore

    @abstractmethod
    async def dump(self, strip: bool = False) -> bytes:
        """Reads the entire contents of the memory and returns it as a bytes
        object.

        Parameters:
            strip: whether to strip trailing zero bytes from the result
        """
        raise NotImplementedError

    @abstractmethod
    async def read(self, addr: int, length: int) -> bytes:
        """Reads a given number of bytes from the given address.

        Parameters:
            addr: the address to read from
            length: the number of bytes to read
        """
        raise NotImplementedError

    @abstractproperty
    def size(self) -> int:
        """Returns the size of the memory that this handler handles."""
        raise NotImplementedError

    @abstractproperty
    def type(self) -> int:
        """Returns the type of the memory that this handler handles."""
        raise NotImplementedError

    @abstractmethod
    async def write(self, addr: int, data: bytes) -> None:
        """Writes some data to the given address.

        Parameters:
            addr: the address to read from
            length: the number of bytes to read
        """
        raise NotImplementedError


class MemoryHandlerBase(MemoryHandler):
    """Base implementation of a memory handler."""

    _addressing_struct: ClassVar[Struct] = Struct("<BI")

    def __init__(self, element: MemoryElement, owner: Crazyflie):
        """Constructor.

        Parameters:
            element: the memory element object that contains the type, size,
                address and index of the memory managed by this handler
            owner: the Crazyflie that owns this handler
        """
        self._crazyflie = owner
        self._element = element

    async def dump(self, strip: bool = False) -> bytes:
        result = await self.read(0, self.size)
        return result.rstrip(b"\x00") if strip else result

    async def read(self, addr: int, length: int) -> bytes:
        chunks = []
        for start, size in chunkify(
            addr, length, step=MemoryHandler.MAX_READ_REQUEST_LENGTH
        ):
            chunk, status = await self._read_chunk(start, size)
            if status == 0:
                chunks.append(chunk)
            else:
                raise IOError(
                    status,
                    "Read request returned error code {0} ({1})".format(
                        status, error_to_string(status)
                    ),
                )
        return b"".join(chunks)

    @property
    def size(self) -> int:
        return self._element.size

    @property
    def type(self) -> int:
        return self._element.type

    async def write(self, addr: int, data: bytes) -> None:
        for start, size in chunkify(
            0, len(data), step=MemoryHandler.MAX_WRITE_REQUEST_LENGTH
        ):
            status = await self._write_chunk(addr + start, data[start : (start + size)])
            if status:
                raise IOError(
                    status,
                    "Write request returned error code {0} ({1})".format(
                        status, error_to_string(status)
                    ),
                )

    async def _read_chunk(self, addr: int, length: int) -> Tuple[bytes, int]:
        """Reads a chunk of data that fits into a single packet, starting
        from the given address.

        Parameters:
            addr: the address to start the read operation from
            length: the number of bytes to read; must be less than
                `MemoryHandler.MAX_READ_REQUEST_LENGTH`.

        Returns:
            the data that was read and the status code sent by the Crazyflie,
            in a tuple. A zero status code means that the read operation was
            successful.
        """
        response = await self._crazyflie.run_command(
            port=CRTPPort.MEMORY,
            channel=MemoryChannel.READ,
            command=self._addressing_struct.pack(self._element.index, addr),
            data=(length,),
        )
        return (response[1:], response[0]) if response else (b"", ENODATA)

    async def _write_chunk(self, addr: int, data: bytes) -> int:
        """Writes a chunk of data that fits into a single packet, starting
        from the given address.

        Parameters:
            addr: the address to write the data to
            data: the data to write; must be shorter than
                `MemoryHandler.MAX_WRITE_REQUEST_LENGTH`.

        Returns:
            the status code sent by the Crazyflie; zero means that the write
            was successful
        """
        response = await self._crazyflie.run_command(
            port=CRTPPort.MEMORY,
            channel=MemoryChannel.WRITE,
            command=self._addressing_struct.pack(self._element.index, addr),
            data=data,
        )
        return response[0] if response else ENODATA


class Memory:
    """Class representing the handler of messages related to the memory
    subsystem of a Crazyflie instance.
    """

    _crazyflie: Crazyflie
    _handlers: Optional[List[MemoryHandler]]

    def __init__(self, crazyflie: Crazyflie):
        """Constructor.

        Parameters:
            crazyflie: the Crazyflie for which we need to handle the memory
                subsystem related messages
        """
        self._crazyflie = crazyflie
        self._handlers = None

    async def find(self, type: MemoryType) -> MemoryHandler:
        """Finds the first memory element with the given type.

        Parameters:
            type: the type of the memory element to look for

        Returns:
            a handler object that can be used to read from and write to the
            given memory

        Raises:
            ValueError: if there is no such memory element
        """
        await self.validate()
        assert self._handlers is not None
        for handler in self._handlers:
            if handler.type == type:
                return handler
        raise ValueError("no memory matching type {0!r}".format(type))

    async def find_all(self, type: MemoryType) -> List[MemoryHandler]:
        """Finds all memory elements with the given type.

        Parameters:
            type: the type of the memory element to look for

        Yields:
            handler objects for all memory elements that have the given type.
            The handler objects can be used to read from and write to the
            corresponding memory element.
        """
        await self.validate()
        assert self._handlers is not None
        return [handler for handler in self._handlers if handler.type == type]

    async def find_eeprom(self) -> MemoryHandler:
        """Shortcut to find the memory handler for the internal EEPROM of the
        Crazyflie where the basic configuration settings are stored.
        """
        return await self.find(MemoryType.I2C)

    async def read(self, type: MemoryType, addr: int, length: int) -> bytes:
        """Shortcut to read the given number of bytes from the given address
        of the first memory element of the given type.

        Parameters:
            type: the type of the memory element to look for
            addr: the address to read from
            length: the number of bytes to read
        """
        handler = await self.find(type)
        return await handler.read(addr, length)

    async def validate(self):
        """Ensures that the basic information about the memories on the Crazyflie
        are downloaded.
        """
        if self._handlers is not None:
            return

        self._handlers = await self._validate()

    async def write(self, type: MemoryType, addr: int, data: bytes) -> None:
        """Shortcut to write the given data to the given address of the first
        memory element of the given type.

        Parameters:
            type: the type of the memory element to write to
            addr: the address to write to
            data: the data to write
        """
        handler = await self.find(type)
        return await handler.write(addr, data)

    async def write_with_checksum(
        self,
        type: MemoryType,
        addr: int,
        data: bytes,
        *,
        only_if_changed: bool = False,
        checksum: Callable[[bytes], bytes] = crc32,
    ) -> int:
        """Writes some data to the given address of the first memory element of
        the given type, _prepended by a checksum_.

        The primary purpose of this function is to prevent spending time with
        writing some data to some place in the Crazyflie memory if the same data
        has been written before. See the documentation of the
        `write_with_checksum()` function for more details; this method is just
        a convenience wrapper around it.

        Parameters:
            type: the type of the memory element to write to
            addr: the address to write to. The first few bytes will contain the
                checksum; the real data will be written _after_ the checksum.
                The length of the checksum will be returned from the function.
            data: the data to write
            only_if_changed: whether to write the data only if we detect that
                the checksum in front of the data is not identical to the
                expected checksum. When this parameter is False (which is the
                default), the data will be written unconditionally.
            checksum: the checksum function. This function must take the data to
                write and return a bytes object of _fixed_ length that contains
                the checksum of the data.

        Returns:
            the number of checksum bytes to skip if we want to read the data
            only, without its checksum
        """
        handler = await self.find(type)
        return await write_with_checksum(
            handler, addr, data, only_if_changed=only_if_changed, checksum=checksum
        )

    async def _get_memory_details(self, index: int) -> MemoryElement:
        """Retrieves detailed information about a single memory with the given
        index.

        Parameters:
            index: the index of the memory for which we need to retrieve its
            details

        Returns:
            the parsed details of the memory

        Raises:
            IndexError: if there is no such memory with the given index
            ValueError: if the response from the Crazyflie cannot be parsed
        """
        response = await self._crazyflie.run_command(
            port=CRTPPort.MEMORY,
            channel=MemoryChannel.INFO,
            command=(MemoryInfoCommand.GET_DETAILS, index),
        )
        if response:
            return MemoryElement.from_bytes(index, response)
        else:
            raise IndexError("memory index out of range")

    async def _get_number_of_memories(self) -> int:
        """Returns the number of memories present on the Crazyflie."""
        response = await self._crazyflie.run_command(
            port=CRTPPort.MEMORY,
            channel=MemoryChannel.INFO,
            command=MemoryInfoCommand.GET_NUMBER_OF_MEMORIES,
        )
        return response[0]

    async def _validate(self) -> List[MemoryHandler]:
        """Downloads the basic information about the memories on the Crazyflie."""
        num_memories = await self._get_number_of_memories()
        memories = [await self._get_memory_details(i) for i in range(num_memories)]
        return [
            MemoryHandler.for_element(memory, owner=self._crazyflie)
            for memory in memories
        ]


async def write_with_checksum(
    handler: MemoryHandler,
    addr: int,
    data: bytes,
    *,
    only_if_changed: bool = False,
    checksum: Callable[[bytes], bytes] = crc32,
) -> int:
    """Writes some data to the given address, _prepended by a checksum_.

    The primary purpose of this function is to prevent spending time with
    writing some data to some place in the Crazyflie memory if the same data
    has been written before. This is achieved by prepending the data with a
    checksum of fixed length. During subsequent write attempts, one can read
    the checksum first and compare it with the expected checksum of the
    data; if the two are the same, one can simply assume that the data has
    already been written and skip writing it again.

    Implementation note: since the Crazyflie memory segments are initialized
    to all-zeros after powerup, we must ensure that the checksum that we
    write to the Crazyflie memory is never zero. When the checksum function
    supplied by the user returns a checksum where all bytes are zeros, the
    checksum will be replaced by a placeholder value.

    Parameters:
        addr: the address to write to. The first few bytes will contain the
            checksum; the real data will be written _after_ the checksum.
            The length of the checksum will be returned from the function.
        data: the data to write
        only_if_changed: whether to write the data only if we detect that
            the checksum in front of the data is not identical to the
            expected checksum. When this parameter is False (which is the
            default), the data will be written unconditionally.
        checksum: the checksum function. This function must take the data to
            write and return a bytes object of _fixed_ length that contains
            the checksum of the data.

    Returns:
        the number of checksum bytes to skip if we want to read the data
        only, without its checksum
    """
    expected_chksum = checksum(data)
    chksum_length = len(expected_chksum)

    if not only_if_changed:
        need_to_write = True
    else:
        observed_chksum = await handler.read(addr, chksum_length)
        need_to_write = observed_chksum != expected_chksum

    if need_to_write:
        zeros = bytes([0] * chksum_length)
        await handler.write(addr, zeros)
        await handler.write(addr + chksum_length, data)
        await handler.write(addr, expected_chksum)

    return chksum_length



from typing import Callable

from aiocflib.utils.registry import Registry

from .base import CRTPDriver

__all__ = ("DriverRegistry", "find", "register")


#: Type specification for CRTP driver factories
CRTPDriverFactory = Callable[[], CRTPDriver]

#: Mapping that maps names to the corresponding CRTPDriver classes
DriverRegistry = Registry()  # type: Registry[CRTPDriverFactory]

find = DriverRegistry.find
"""Returns the registered CRTP driver factory corresponding to the given
name.

Parameters:
    name: the name to which the driver factory is registered

Raises:
    KeyError: if there is no registered CRTP driver factory for the given
        name

Returns:
    the driver factory corresponding to the name
"""

register = DriverRegistry.register
"""Class decorator factory that returns a decorator that registers a class
as a CRTP driver with the given name.

Parameters:
    name: the name to which the driver factory is registered

Returns:
    an appropriate decorator that can then be applied to a CRTPDriver
    subclass
"""
